"""Data classes for Resource Usage History repository layer."""

from __future__ import annotations

import uuid
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import date, datetime
from typing import Any

import sqlalchemy as sa

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.errors.resource import DomainNotFound, ProjectNotFound, ScalingGroupNotFound
from ai.backend.manager.errors.user import UserNotFound
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.group import GroupRow
from ai.backend.manager.models.resource_usage_history import (
    DomainUsageBucketRow,
    KernelUsageRecordRow,
    ProjectUsageBucketRow,
    UserUsageBucketRow,
)
from ai.backend.manager.models.scaling_group import ScalingGroupRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base import ExistenceCheck, QueryCondition, SearchScope


@dataclass(frozen=True)
class KernelUsageRecordData:
    """Kernel usage record data (per-period usage slice)."""

    id: uuid.UUID
    kernel_id: uuid.UUID
    session_id: uuid.UUID
    user_uuid: uuid.UUID
    project_id: uuid.UUID
    domain_name: str
    resource_group: str
    period_start: datetime
    period_end: datetime
    resource_usage: ResourceSlot

    @classmethod
    def from_row(cls, row: KernelUsageRecordRow) -> KernelUsageRecordData:
        """Create KernelUsageRecordData from a KernelUsageRecordRow."""
        return cls(
            id=row.id,
            kernel_id=row.kernel_id,
            session_id=row.session_id,
            user_uuid=row.user_uuid,
            project_id=row.project_id,
            domain_name=row.domain_name,
            resource_group=row.resource_group,
            period_start=row.period_start,
            period_end=row.period_end,
            resource_usage=row.resource_usage,
        )


@dataclass(frozen=True)
class DomainUsageBucketData:
    """Domain usage bucket data (period-based aggregation)."""

    id: uuid.UUID
    domain_name: str
    resource_group: str
    period_start: date
    period_end: date
    decay_unit_days: int
    resource_usage: ResourceSlot
    capacity_snapshot: ResourceSlot
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_row(cls, row: DomainUsageBucketRow) -> DomainUsageBucketData:
        """Create DomainUsageBucketData from a DomainUsageBucketRow."""
        return cls(
            id=row.id,
            domain_name=row.domain_name,
            resource_group=row.resource_group,
            period_start=row.period_start,
            period_end=row.period_end,
            decay_unit_days=row.decay_unit_days,
            resource_usage=row.resource_usage,
            capacity_snapshot=row.capacity_snapshot,
            created_at=row.created_at,
            updated_at=row.updated_at,
        )


@dataclass(frozen=True)
class ProjectUsageBucketData:
    """Project usage bucket data (period-based aggregation)."""

    id: uuid.UUID
    project_id: uuid.UUID
    domain_name: str
    resource_group: str
    period_start: date
    period_end: date
    decay_unit_days: int
    resource_usage: ResourceSlot
    capacity_snapshot: ResourceSlot
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_row(cls, row: ProjectUsageBucketRow) -> ProjectUsageBucketData:
        """Create ProjectUsageBucketData from a ProjectUsageBucketRow."""
        return cls(
            id=row.id,
            project_id=row.project_id,
            domain_name=row.domain_name,
            resource_group=row.resource_group,
            period_start=row.period_start,
            period_end=row.period_end,
            decay_unit_days=row.decay_unit_days,
            resource_usage=row.resource_usage,
            capacity_snapshot=row.capacity_snapshot,
            created_at=row.created_at,
            updated_at=row.updated_at,
        )


@dataclass(frozen=True)
class UserUsageBucketData:
    """User usage bucket data (period-based aggregation)."""

    id: uuid.UUID
    user_uuid: uuid.UUID
    project_id: uuid.UUID
    domain_name: str
    resource_group: str
    period_start: date
    period_end: date
    decay_unit_days: int
    resource_usage: ResourceSlot
    capacity_snapshot: ResourceSlot
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_row(cls, row: UserUsageBucketRow) -> UserUsageBucketData:
        """Create UserUsageBucketData from a UserUsageBucketRow."""
        return cls(
            id=row.id,
            user_uuid=row.user_uuid,
            project_id=row.project_id,
            domain_name=row.domain_name,
            resource_group=row.resource_group,
            period_start=row.period_start,
            period_end=row.period_end,
            decay_unit_days=row.decay_unit_days,
            resource_usage=row.resource_usage,
            capacity_snapshot=row.capacity_snapshot,
            created_at=row.created_at,
            updated_at=row.updated_at,
        )


@dataclass(frozen=True)
class KernelUsageRecordSearchResult:
    """Search result with pagination info for kernel usage records."""

    items: list[KernelUsageRecordData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass(frozen=True)
class DomainUsageBucketSearchResult:
    """Search result with pagination info for domain usage buckets."""

    items: list[DomainUsageBucketData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass(frozen=True)
class ProjectUsageBucketSearchResult:
    """Search result with pagination info for project usage buckets."""

    items: list[ProjectUsageBucketData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass(frozen=True)
class UserUsageBucketSearchResult:
    """Search result with pagination info for user usage buckets."""

    items: list[UserUsageBucketData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


# SearchScope classes for scoped usage bucket APIs


@dataclass(frozen=True)
class DomainUsageBucketSearchScope(SearchScope):
    """Scope for domain usage bucket queries."""

    resource_group: str
    domain_name: str

    def to_condition(self) -> QueryCondition:
        resource_group = self.resource_group
        domain_name = self.domain_name

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                DomainUsageBucketRow.resource_group == resource_group,
                DomainUsageBucketRow.domain_name == domain_name,
            )

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[Any]]:
        return [
            ExistenceCheck(
                column=ScalingGroupRow.name,
                value=self.resource_group,
                error=ScalingGroupNotFound(self.resource_group),
            ),
            ExistenceCheck(
                column=DomainRow.name,
                value=self.domain_name,
                error=DomainNotFound(self.domain_name),
            ),
        ]


@dataclass(frozen=True)
class ProjectUsageBucketSearchScope(SearchScope):
    """Scope for project usage bucket queries."""

    resource_group: str
    domain_name: str
    project_id: uuid.UUID

    def to_condition(self) -> QueryCondition:
        resource_group = self.resource_group
        domain_name = self.domain_name
        project_id = self.project_id

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                ProjectUsageBucketRow.resource_group == resource_group,
                ProjectUsageBucketRow.domain_name == domain_name,
                ProjectUsageBucketRow.project_id == project_id,
            )

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[Any]]:
        return [
            ExistenceCheck(
                column=ScalingGroupRow.name,
                value=self.resource_group,
                error=ScalingGroupNotFound(self.resource_group),
            ),
            ExistenceCheck(
                column=DomainRow.name,
                value=self.domain_name,
                error=DomainNotFound(self.domain_name),
            ),
            ExistenceCheck(
                column=GroupRow.id,
                value=self.project_id,
                error=ProjectNotFound(extra_data={"project_id": str(self.project_id)}),
            ),
        ]


@dataclass(frozen=True)
class UserUsageBucketSearchScope(SearchScope):
    """Scope for user usage bucket queries."""

    resource_group: str
    domain_name: str
    project_id: uuid.UUID
    user_uuid: uuid.UUID

    def to_condition(self) -> QueryCondition:
        resource_group = self.resource_group
        domain_name = self.domain_name
        project_id = self.project_id
        user_uuid = self.user_uuid

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                UserUsageBucketRow.resource_group == resource_group,
                UserUsageBucketRow.domain_name == domain_name,
                UserUsageBucketRow.project_id == project_id,
                UserUsageBucketRow.user_uuid == user_uuid,
            )

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[Any]]:
        return [
            ExistenceCheck(
                column=ScalingGroupRow.name,
                value=self.resource_group,
                error=ScalingGroupNotFound(self.resource_group),
            ),
            ExistenceCheck(
                column=DomainRow.name,
                value=self.domain_name,
                error=DomainNotFound(self.domain_name),
            ),
            ExistenceCheck(
                column=GroupRow.id,
                value=self.project_id,
                error=ProjectNotFound(extra_data={"project_id": str(self.project_id)}),
            ),
            ExistenceCheck(
                column=UserRow.uuid,
                value=self.user_uuid,
                error=UserNotFound(extra_data={"user_uuid": str(self.user_uuid)}),
            ),
        ]



"""
Agent RPC connection pool.
"""

from __future__ import annotations

import asyncio
import logging
import time
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING

import zmq
from aiotools import cancel_and_wait
from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport

from ai.backend.common import msgpack
from ai.backend.common.auth import ManagerAuthHandler
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.agent_cache import PeerInvoker
from ai.backend.manager.errors.agent import AgentConnectionUnavailable

from .client import AgentClient
from .types import AgentPoolSpec

if TYPE_CHECKING:
    from ai.backend.manager.agent_cache import AgentRPCCache, PeerInvoker

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

# Connection-related error types
CONNECTION_ERRORS = (
    ConnectionError,
    asyncio.TimeoutError,
    OSError,
)


@dataclass(slots=True)
class _CachedEntry:
    """Internal cache entry for pool management."""

    client: AgentClient
    is_healthy: bool = True
    failure_count: int = 0
    unhealthy_since: float | None = None


class AgentClientPool:
    """
    Agent RPC connection pool.

    Connection state management:
    1. Periodic health check: ping to verify connection, delete if unrecoverable
    2. Usage-time failure tracking: increment failure_count on connection error during acquire()

    Separation of concerns:
    - acquire(): client acquisition only (raise exception if unhealthy)
    - health_check_loop: connection state management and deletion
    """

    _agent_cache: AgentRPCCache
    _spec: AgentPoolSpec
    _entries: dict[AgentId, _CachedEntry]
    _lock: asyncio.Lock
    _health_check_task: asyncio.Task[None]

    def __init__(
        self,
        agent_cache: AgentRPCCache,
        spec: AgentPoolSpec,
    ) -> None:
        self._agent_cache = agent_cache
        self._spec = spec
        self._entries = {}
        self._lock = asyncio.Lock()

        # Start background task in constructor
        self._health_check_task = asyncio.create_task(
            self._health_check_loop(),
            name="AgentClientPool._health_check_loop",
        )

    async def close(self) -> None:
        """Close the pool."""
        await cancel_and_wait(self._health_check_task)

        async with self._lock:
            for entry in self._entries.values():
                await entry.client.close()
            self._entries.clear()

    @asynccontextmanager
    async def acquire(self, agent_id: AgentId) -> AsyncIterator[AgentClient]:
        """
        Acquire an agent client for use.

        On connection error during use, increment failure count and
        mark as unhealthy if threshold is exceeded.
        Unhealthy connections raise an exception (deletion is handled by health_check_loop).

        Business logic errors do not affect failure count.
        """
        client = await self._get_or_create(agent_id)
        try:
            yield client
        except CONNECTION_ERRORS:
            self._record_failure(agent_id)
            raise
        except Exception:
            # Non-connection errors don't increment failure count
            raise
        else:
            self._record_success(agent_id)

    async def _get_or_create(self, agent_id: AgentId) -> AgentClient:
        """Return healthy client, create if not exists."""
        async with self._lock:
            entry = self._entries.get(agent_id)

            # Raise exception if unhealthy (deletion is handled by health_check_loop)
            if entry is not None and not entry.is_healthy:
                raise AgentConnectionUnavailable(agent_id, "connection unhealthy")

            # Create new entry if not exists
            if entry is None:
                entry = await self._create_entry(agent_id)
                self._entries[agent_id] = entry

            return entry.client

    async def _create_entry(self, agent_id: AgentId) -> _CachedEntry:
        """Create new entry (called within lock)."""
        try:
            agent_addr, agent_public_key = await self._agent_cache.get_rpc_args(agent_id)
        except ValueError as e:
            raise AgentConnectionUnavailable(agent_id, str(e)) from e

        # Set up auth handler
        if agent_public_key:
            auth_handler = ManagerAuthHandler(
                "local",
                agent_public_key,
                self._agent_cache.manager_public_key,
                self._agent_cache.manager_secret_key,
            )
        else:
            auth_handler = None

        # Calculate keepalive settings
        keepalive_retry_count = 3
        keepalive_timeout = self._agent_cache.rpc_keepalive_timeout
        keepalive_interval = keepalive_timeout // keepalive_retry_count
        if keepalive_interval < 2:
            keepalive_interval = 2

        peer: PeerInvoker = self._create_peer(
            agent_addr,
            auth_handler,
            keepalive_timeout,
            keepalive_interval,
            keepalive_retry_count,
        )

        client = AgentClient(peer, agent_id)
        try:
            await client.connect()
        except Exception as e:
            raise AgentConnectionUnavailable(agent_id, str(e)) from e

        return _CachedEntry(
            client=client,
            is_healthy=True,
            failure_count=0,
            unhealthy_since=None,
        )

    def _create_peer(
        self,
        agent_addr: str,
        auth_handler: ManagerAuthHandler | None,
        keepalive_timeout: int,
        keepalive_interval: int,
        keepalive_retry_count: int,
    ) -> PeerInvoker:
        """Create a new PeerInvoker instance."""
        return PeerInvoker(
            connect=ZeroMQAddress(agent_addr),
            transport=ZeroMQRPCTransport,
            authenticator=auth_handler,
            transport_opts={
                "zsock_opts": {
                    zmq.TCP_KEEPALIVE: 1,
                    zmq.TCP_KEEPALIVE_IDLE: keepalive_timeout,
                    zmq.TCP_KEEPALIVE_INTVL: keepalive_interval,
                    zmq.TCP_KEEPALIVE_CNT: keepalive_retry_count,
                },
            },
            serializer=msgpack.packb,
            deserializer=msgpack.unpackb,
        )

    def _record_failure(self, agent_id: AgentId) -> None:
        """Record connection error and mark unhealthy if threshold exceeded."""
        entry = self._entries.get(agent_id)
        if entry is None:
            return

        entry.failure_count += 1
        if entry.failure_count >= self._spec.failure_threshold:
            entry.is_healthy = False
            if entry.unhealthy_since is None:
                entry.unhealthy_since = time.perf_counter()
            log.debug(
                "Agent {} marked unhealthy after {} connection failures",
                agent_id,
                entry.failure_count,
            )

    def _record_success(self, agent_id: AgentId) -> None:
        """Reset state on success."""
        entry = self._entries.get(agent_id)
        if entry is not None:
            entry.failure_count = 0
            entry.is_healthy = True
            entry.unhealthy_since = None

    async def _health_check_loop(self) -> None:
        """Periodically check all connection health."""
        while True:
            await asyncio.sleep(self._spec.health_check_interval)
            await self._check_all_health()

    async def _check_all_health(self) -> None:
        """Check health of all connections (using asyncio.gather)."""
        async with self._lock:
            entries = list(self._entries.items())

        if not entries:
            return

        await asyncio.gather(
            *[self._check_one_health(agent_id, entry) for agent_id, entry in entries],
            return_exceptions=True,
        )

    async def _check_one_health(self, agent_id: AgentId, entry: _CachedEntry) -> None:
        """Check single connection health and delete if unrecoverable."""
        try:
            async with asyncio.timeout(5.0):
                await entry.client.ping()
            # Ping success → recover
            entry.is_healthy = True
            entry.failure_count = 0
            entry.unhealthy_since = None
        except Exception:
            # Ping failure → mark unhealthy
            entry.is_healthy = False
            if entry.unhealthy_since is None:
                entry.unhealthy_since = time.perf_counter()
                log.debug("Health check failed for agent {}", agent_id)

        # Delete if recovery_timeout exceeded
        if (
            not entry.is_healthy
            and entry.unhealthy_since is not None
            and time.perf_counter() - entry.unhealthy_since > self._spec.recovery_timeout
        ):
            async with self._lock:
                if agent_id in self._entries:
                    await entry.client.close()
                    del self._entries[agent_id]
            log.info(
                "Removed unrecoverable connection for agent {} after {}s",
                agent_id,
                self._spec.recovery_timeout,
            )



from dataclasses import dataclass

from sqlalchemy.sql.elements import ColumnElement


@dataclass(frozen=True)
class SearchResult[T]:
    items: list[T]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass
class StringFilterData:
    """Data class that corresponds 1:1 with StringFilter for API requests."""

    contains: str | None = None
    starts_with: str | None = None
    ends_with: str | None = None
    equals: str | None = None
    not_equals: str | None = None
    i_contains: str | None = None
    i_starts_with: str | None = None
    i_ends_with: str | None = None
    i_equals: str | None = None
    i_not_equals: str | None = None

    def apply_to_column(self, column: ColumnElement[str]) -> ColumnElement[bool] | None:
        """Apply this string filter to a SQLAlchemy column and return the condition.

        Args:
            column: SQLAlchemy column to apply the filter to

        Returns:
            SQLAlchemy condition expression or None if no filter is set
        """
        if self.equals is not None:
            return column == self.equals
        if self.i_equals is not None:
            return column.ilike(self.i_equals)
        if self.not_equals is not None:
            return column != self.not_equals
        if self.i_not_equals is not None:
            return ~column.ilike(self.i_not_equals)
        if self.starts_with is not None:
            return column.like(f"{self.starts_with}%")
        if self.i_starts_with is not None:
            return column.ilike(f"{self.i_starts_with}%")
        if self.ends_with is not None:
            return column.like(f"%{self.ends_with}")
        if self.i_ends_with is not None:
            return column.ilike(f"%{self.i_ends_with}")
        if self.contains is not None:
            return column.like(f"%{self.contains}%")
        if self.i_contains is not None:
            return column.ilike(f"%{self.i_contains}%")

        return None


@dataclass
class IntFilterData:
    """Data class that corresponds 1:1 with IntFilter for API requests."""

    equals: int | None = None
    not_equals: int | None = None
    greater_than: int | None = None
    greater_than_or_equal: int | None = None
    less_than: int | None = None
    less_than_or_equal: int | None = None

    def apply_to_column(self, column: ColumnElement[int]) -> ColumnElement[bool] | None:
        """Apply this int filter to a SQLAlchemy column and return the condition.

        Args:
            column: SQLAlchemy column to apply the filter to

        Returns:
            SQLAlchemy condition expression or None if no filter is set
        """
        if self.equals is not None:
            return column == self.equals
        if self.not_equals is not None:
            return column != self.not_equals
        if self.greater_than is not None:
            return column > self.greater_than
        if self.greater_than_or_equal is not None:
            return column >= self.greater_than_or_equal
        if self.less_than is not None:
            return column < self.less_than
        if self.less_than_or_equal is not None:
            return column <= self.less_than_or_equal

        return None



"""Handler for starting sessions."""

from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING

from ai.backend.common.types import AccessKey
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus, StatusTransitions, TransitionStatus
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.base import BatchQuerier, NoPagination
from ai.backend.manager.repositories.scheduler import SchedulerRepository
from ai.backend.manager.repositories.scheduler.options import SessionConditions
from ai.backend.manager.sokovan.data import SessionWithKernels
from ai.backend.manager.sokovan.scheduler.handlers.base import SessionLifecycleHandler
from ai.backend.manager.sokovan.scheduler.results import (
    SessionExecutionResult,
    SessionTransitionInfo,
)

if TYPE_CHECKING:
    from ai.backend.manager.sokovan.scheduler.launcher.launcher import SessionLauncher

log = BraceStyleAdapter(logging.getLogger(__name__))


class StartSessionsLifecycleHandler(SessionLifecycleHandler):
    """Handler for starting sessions that passed precondition checks.

    Following the DeploymentCoordinator pattern:
    - Coordinator queries sessions with PREPARED status (provides HandlerSessionData)
    - Handler queries additional data (SessionDataForStart + ImageConfigData) via Repository
    - Handler starts kernels on agents via Launcher
    - Coordinator updates sessions to CREATING status and broadcasts events
    """

    def __init__(
        self,
        launcher: SessionLauncher,
        repository: SchedulerRepository,
    ) -> None:
        self._launcher = launcher
        self._repository = repository

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "start-sessions"

    @classmethod
    def target_statuses(cls) -> list[SessionStatus]:
        """Sessions in PREPARED state."""
        return [SessionStatus.PREPARED]

    @classmethod
    def target_kernel_statuses(cls) -> list[KernelStatus] | None:
        """Include sessions where kernels are in PREPARED status."""
        return [KernelStatus.PREPARED]

    @classmethod
    def status_transitions(cls) -> StatusTransitions:
        """Define state transitions for start sessions handler (BEP-1030).

        - success: Session/kernel → CREATING
        - need_retry: None (stays PREPARED)
        - expired: Session/kernel → PENDING (re-scheduling after timeout)
        - give_up: None (container creation is time-based, only timeout applies)
        """
        return StatusTransitions(
            success=TransitionStatus(
                session=SessionStatus.CREATING,
                kernel=KernelStatus.CREATING,
            ),
            need_retry=None,
            expired=TransitionStatus(
                session=SessionStatus.PENDING,
                kernel=KernelStatus.PENDING,
            ),
            give_up=None,
        )

    @property
    def lock_id(self) -> LockID | None:
        """Lock for operations targeting PREPARED sessions transitioning to CREATING."""
        return LockID.LOCKID_SOKOVAN_TARGET_CREATING

    async def execute(
        self,
        _scaling_group: str,
        sessions: Sequence[SessionWithKernels],
    ) -> SessionExecutionResult:
        """Start kernels on agents for PREPARED sessions.

        The coordinator provides SessionWithKernels data.
        This handler:
        1. Extracts session IDs from SessionWithKernels
        2. Queries Repository for additional data (SessionDataForStart + ImageConfigData)
        3. Starts kernels on agents via Launcher
        """
        result = SessionExecutionResult()

        if not sessions:
            return result

        # Extract session IDs from SessionWithKernels
        session_ids = [s.session_info.identity.id for s in sessions]

        # Query Repository for additional data needed by Launcher
        # Use search_sessions_with_kernels_and_user to get user info for session start
        querier = BatchQuerier(
            pagination=NoPagination(),
            conditions=[SessionConditions.by_ids(session_ids)],
        )
        sessions_data = await self._repository.search_sessions_with_kernels_and_user(querier)

        # Start kernels on agents via Launcher
        # Note: RecorderContext is handled inside Launcher
        await self._launcher.start_sessions_for_handler(
            sessions_data.sessions,
            sessions_data.image_configs,
        )

        # Mark all sessions as success for status transition
        for session in sessions:
            session_info = session.session_info
            result.successes.append(
                SessionTransitionInfo(
                    session_id=session_info.identity.id,
                    from_status=session_info.lifecycle.status,
                    reason="triggered-by-scheduler",
                    creation_id=session_info.identity.creation_id,
                    access_key=AccessKey(session_info.metadata.access_key),
                )
            )

        return result



"""Fair Share GQL types package."""

from .common import (
    FairShareCalculationSnapshotGQL,
    FairShareSpecGQL,
    ResourceSlotEntryGQL,
    ResourceSlotGQL,
    ResourceWeightEntryInputGQL,
)
from .domain import (
    BulkUpsertDomainFairShareWeightInput,
    BulkUpsertDomainFairShareWeightPayload,
    DomainFairShareConnection,
    DomainFairShareDomainNestedFilter,
    DomainFairShareEdge,
    DomainFairShareFilter,
    DomainFairShareGQL,
    DomainFairShareOrderBy,
    DomainFairShareOrderField,
    DomainWeightInputItem,
    UpsertDomainFairShareWeightInput,
    UpsertDomainFairShareWeightPayload,
)
from .project import (
    BulkUpsertProjectFairShareWeightInput,
    BulkUpsertProjectFairShareWeightPayload,
    ProjectFairShareConnection,
    ProjectFairShareEdge,
    ProjectFairShareFilter,
    ProjectFairShareGQL,
    ProjectFairShareOrderBy,
    ProjectFairShareOrderField,
    ProjectFairShareProjectNestedFilter,
    ProjectFairShareTypeEnum,
    ProjectFairShareTypeEnumFilter,
    ProjectWeightInputItem,
    UpsertProjectFairShareWeightInput,
    UpsertProjectFairShareWeightPayload,
)
from .user import (
    BulkUpsertUserFairShareWeightInput,
    BulkUpsertUserFairShareWeightPayload,
    UpsertUserFairShareWeightInput,
    UpsertUserFairShareWeightPayload,
    UserFairShareConnection,
    UserFairShareEdge,
    UserFairShareFilter,
    UserFairShareGQL,
    UserFairShareOrderBy,
    UserFairShareOrderField,
    UserFairShareUserNestedFilter,
    UserWeightInputItem,
)

__all__ = [
    # Common
    "ResourceSlotEntryGQL",
    "ResourceSlotGQL",
    "ResourceWeightEntryInputGQL",
    "FairShareSpecGQL",
    "FairShareCalculationSnapshotGQL",
    # Domain
    "DomainFairShareGQL",
    "DomainFairShareConnection",
    "DomainFairShareDomainNestedFilter",
    "DomainFairShareEdge",
    "DomainFairShareFilter",
    "DomainFairShareOrderField",
    "DomainFairShareOrderBy",
    "UpsertDomainFairShareWeightInput",
    "UpsertDomainFairShareWeightPayload",
    "DomainWeightInputItem",
    "BulkUpsertDomainFairShareWeightInput",
    "BulkUpsertDomainFairShareWeightPayload",
    # Project
    "ProjectFairShareGQL",
    "ProjectFairShareConnection",
    "ProjectFairShareEdge",
    "ProjectFairShareFilter",
    "ProjectFairShareOrderField",
    "ProjectFairShareOrderBy",
    "ProjectFairShareProjectNestedFilter",
    "ProjectFairShareTypeEnum",
    "ProjectFairShareTypeEnumFilter",
    "UpsertProjectFairShareWeightInput",
    "UpsertProjectFairShareWeightPayload",
    "ProjectWeightInputItem",
    "BulkUpsertProjectFairShareWeightInput",
    "BulkUpsertProjectFairShareWeightPayload",
    # User
    "UserFairShareGQL",
    "UserFairShareConnection",
    "UserFairShareEdge",
    "UserFairShareFilter",
    "UserFairShareOrderField",
    "UserFairShareOrderBy",
    "UserFairShareUserNestedFilter",
    "UpsertUserFairShareWeightInput",
    "UpsertUserFairShareWeightPayload",
    "UserWeightInputItem",
    "BulkUpsertUserFairShareWeightInput",
    "BulkUpsertUserFairShareWeightPayload",
]



"""Migrate container registry config storage from `Etcd` to `PostgreSQL`

Revision ID: 1d42c726d8a3
Revises: 20218a73401b
Create Date: 2024-03-05 10:36:24.197922

"""

import asyncio
import enum
import json
import logging
import os
import sys
import warnings
from collections.abc import Mapping
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from itertools import groupby
from pathlib import Path
from queue import Queue
from typing import Any, Final, cast

import sqlalchemy as sa
import trafaret as t
from alembic import op
from sqlalchemy.exc import SAWarning
from sqlalchemy.orm import registry

from ai.backend.common import validators as tx
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.manager.config_legacy import load
from ai.backend.manager.models.base import GUID, IDColumn, StrEnumType, convention

# revision identifiers, used by Alembic.
revision = "1d42c726d8a3"
down_revision = "20218a73401b"
branch_labels = None
depends_on = None

logger = logging.getLogger("alembic.runtime.migration")

warnings.filterwarnings(
    "ignore",
    message="This declarative base already contains a class with the same class name and module name as .*",
    category=SAWarning,
)

metadata = sa.MetaData(naming_convention=convention)
mapper_registry = registry(metadata=metadata)
Base: Any = mapper_registry.generate_base()

etcd_container_registry_iv = t.Dict({
    t.Key(""): tx.URL,
    t.Key("type", default="docker"): t.String,
    t.Key("username", default=None): t.Null | t.String,
    t.Key("password", default=None): t.Null | t.String(allow_blank=True),
    t.Key("project", default=None): (
        t.Null | t.List(t.String) | tx.StringList(empty_str_as_empty_list=True)
    ),
    tx.AliasedKey(["ssl_verify", "ssl-verify"], default=True): t.ToBool,
}).allow_extra("*")

ETCD_CONTAINER_REGISTRY_KEY: Final = "config/docker/registry"
ETCD_BACKUP_FILENAME_PATTERN: Final = "backup.etcd.container-registries.{timestamp}.json"


class ContainerRegistryType(enum.StrEnum):
    DOCKER = "docker"
    HARBOR = "harbor"
    HARBOR2 = "harbor2"
    GITHUB = "github"
    GITLAB = "gitlab"
    ECR = "ecr"
    ECR_PUB = "ecr-public"
    LOCAL = "local"


metadata = sa.MetaData(naming_convention=convention)


class ImageType(enum.Enum):
    COMPUTE = "compute"
    SYSTEM = "system"
    SERVICE = "service"


images_table = sa.Table(
    "images",
    metadata,
    sa.Column("id", GUID, primary_key=True, nullable=False),
    sa.Column("name", sa.String, nullable=False, index=True),
    sa.Column("image", sa.String, nullable=False, index=True),
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now(), index=True),
    sa.Column("tag", sa.TEXT),
    sa.Column("registry", sa.String, nullable=False, index=True),
    sa.Column("registry_id", GUID, nullable=True, index=True),
    sa.Column("architecture", sa.String, nullable=False, index=True, server_default="x86_64"),
    sa.Column("config_digest", sa.CHAR(length=72), nullable=False),
    sa.Column("size_bytes", sa.BigInteger, nullable=False),
    sa.Column("is_local", sa.Boolean, nullable=False, server_default=sa.sql.expression.false()),
    sa.Column("type", sa.Enum(ImageType), nullable=False),
    sa.Column("accelerators", sa.String),
    sa.Column("labels", sa.JSON, nullable=False, server_default="{}"),
    sa.Column("resources", sa.JSON, nullable=False, server_default="{}"),
    extend_existing=True,
)


def get_container_registry_row_schema() -> type[Any]:
    class ContainerRegistryRow(Base):  # type: ignore[misc]
        __tablename__ = "container_registries"
        __table_args__ = {"extend_existing": True}
        id = IDColumn()
        url = sa.Column("url", sa.String(length=512), index=True)
        registry_name = sa.Column("registry_name", sa.String(length=50), index=True)
        type: sa.Column[Any] = sa.Column(
            "type",
            StrEnumType(ContainerRegistryType),
            default=ContainerRegistryType.DOCKER,
            server_default=ContainerRegistryType.DOCKER,
            nullable=False,
            index=True,
        )
        project = sa.Column("project", sa.String(length=255), nullable=True)
        username = sa.Column("username", sa.String(length=255), nullable=True)
        password = sa.Column("password", sa.String, nullable=True)
        ssl_verify = sa.Column("ssl_verify", sa.Boolean, server_default=sa.text("true"), index=True)
        is_global = sa.Column("is_global", sa.Boolean, server_default=sa.text("true"), index=True)

    return ContainerRegistryRow


def get_async_etcd() -> AsyncEtcd:
    local_config = load()
    etcd_config = local_config.get("etcd", None)
    assert etcd_config is not None, "etcd configuration is not found"
    etcd_credentials = None
    if local_config["etcd"]["user"]:
        etcd_credentials = {
            "user": local_config["etcd"]["user"],
            "password": local_config["etcd"]["password"],
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: "",
    }
    return AsyncEtcd(
        local_config["etcd"]["addr"],
        local_config["etcd"]["namespace"],
        scope_prefix_map,
        credentials=etcd_credentials,
    )


def delete_old_etcd_container_registries() -> None:
    queue: Queue[Any] = Queue()

    with ThreadPoolExecutor() as executor:

        def delete_etcd_container_registries() -> None:
            async def _delete_container_registries() -> None:
                etcd = get_async_etcd()
                await etcd.delete_prefix(ETCD_CONTAINER_REGISTRY_KEY)

            queue.put(asyncio.run(_delete_container_registries()))

        executor.submit(delete_etcd_container_registries)


def migrate_data_etcd_to_psql() -> None:
    queue: Queue[Any] = Queue()

    with ThreadPoolExecutor() as executor:

        def backup(etcd_container_registries: Mapping[str, Any]) -> None:
            backup_path = Path(os.getenv("BACKEND_ETCD_BACKUP_PATH", "."))
            backup_path /= ETCD_BACKUP_FILENAME_PATTERN.format(timestamp=datetime.now().isoformat())  # noqa: DTZ005
            with backup_path.open("w") as f:
                json.dump(dict(etcd_container_registries), f, indent=4)

        # If there are no container registries, it returns an empty list.
        # If an error occurs while saving backup, it returns error.
        def get_etcd_container_registries(queue: Queue[Any]) -> None:
            async def _get_container_registries() -> Mapping[str, Any] | Exception:
                etcd = get_async_etcd()
                result = await etcd.get_prefix(ETCD_CONTAINER_REGISTRY_KEY)
                try:
                    backup(result)
                except Exception as e:
                    return e
                return result

            queue.put(asyncio.run(_get_container_registries()))

        executor.submit(get_etcd_container_registries, queue)

    maybe_registries = queue.get()

    if isinstance(maybe_registries, Exception):
        err_msg = (
            f"Failed to save etcd container registries backup file., Cause: {maybe_registries}"
        )
        print(
            err_msg,
            file=sys.stderr,
        )
        raise RuntimeError(err_msg) from maybe_registries
    registries = cast(Mapping[str, Any], maybe_registries)

    old_format_container_registries = {
        hostname: etcd_container_registry_iv.check(item) for hostname, item in registries.items()
    }

    input_configs = []
    for hostname, registry_info in old_format_container_registries.items():
        registry_type = ContainerRegistryType(registry_info["type"])

        input_config_template: dict[str, Any] = {
            "registry_name": hostname,  # hostname is changed to registry_name,
            "url": str(registry_info[""]),
            "type": registry_type,
            "username": registry_info.get("username", None),
            "password": registry_info.get("password", None),
            "ssl_verify": registry_info.get("ssl_verify", None),
            "is_global": True,
        }

        # Comma-separated projects should be divided into multiple ContainerRegistryRows.
        if projects := registry_info.get("project"):
            for project in projects:
                input_config = input_config_template.copy()
                input_config["project"] = project
                input_configs.append(input_config)
        else:
            if registry_type == ContainerRegistryType.DOCKER and hostname == "index.docker.io":
                input_config_template["project"] = "library"
            else:
                # ContainerRegistryRow with empty project is allowed only if local registry.
                if registry_type == ContainerRegistryType.LOCAL:
                    input_config_template["project"] = None
                else:
                    raise Exception(
                        f'ContainerRegistryRow.project is required! Please put the value to "{ETCD_CONTAINER_REGISTRY_KEY}/{hostname}/project" before migration.'
                    )

            input_configs.append(input_config_template)

    if not input_configs:
        print("There is no container registries data to migrate in etcd.")
        return

    ContainerRegistryRow = get_container_registry_row_schema()

    db_connection = op.get_bind()
    db_connection.execute(sa.insert(ContainerRegistryRow).values(input_configs))


def revert_data_psql_to_etcd() -> None:
    ContainerRegistryRow = get_container_registry_row_schema()

    db_connection = op.get_bind()

    rows = db_connection.execute(
        sa.select(
            ContainerRegistryRow.url,
            ContainerRegistryRow.registry_name,
            ContainerRegistryRow.type,
            ContainerRegistryRow.project,
            ContainerRegistryRow.username,
            ContainerRegistryRow.password,
            ContainerRegistryRow.ssl_verify,
        )
    ).fetchall()
    items = []

    for url, registry_name, type, project, username, password, ssl_verify in rows:
        item = {
            "": url,
            "type": str(type),
            "hostname": registry_name,  # registry_name should be reverted to hostname,
        }

        if project is not None:
            item["project"] = project
        if username is not None:
            item["username"] = username
        if password is not None:
            item["password"] = password
        if ssl_verify is not None:
            item["ssl_verify"] = ssl_verify

        items.append(item)

    # Records with the same hostname should be grouped together.
    # Note that if there are columns other than the project column that have different values between records,
    # information loss can occur while the reverting process.
    grouped_items = {k: list(v) for k, v in groupby(items, key=lambda x: x["hostname"])}

    def merge_items(items: list[dict[str, Any]]) -> dict[str, Any]:
        for item in items:
            if "project" not in item:
                return items[0]

        projects = [item["project"] for item in items]
        merged_projects = ",".join(projects)

        merged_item = items[0].copy()
        merged_item["project"] = merged_projects

        return merged_item

    merged_items = [merge_items(items) for items in grouped_items.values()]

    def put_etcd_container_registries(merged_items: list[Any], queue: Queue[Any]) -> None:
        etcd = get_async_etcd()
        for item in merged_items:
            hostname = item.pop("hostname")
            asyncio.run(etcd.put_prefix(f"{ETCD_CONTAINER_REGISTRY_KEY}/{hostname}", item))

        queue.put(True)

    queue: Queue[Any] = Queue()

    with ThreadPoolExecutor() as executor:
        executor.submit(put_etcd_container_registries, merged_items, queue)

    queue.get()


def insert_registry_id_to_images() -> None:
    db_connection = op.get_bind()
    ContainerRegistry = get_container_registry_row_schema()

    registry_infos = db_connection.execute(
        sa.select(
            ContainerRegistry.id,
            ContainerRegistry.registry_name,
            ContainerRegistry.project,
        )
    ).fetchall()

    if registry_infos:
        insert_registry_id_query = (
            sa.update(images_table)
            .values(registry_id=sa.bindparam("registry_id"))
            .where(images_table.c.name.startswith(sa.bindparam("registry_name_and_project")))
        )

        query_params = []
        for registry_info in registry_infos:
            registry_id, registry_name, project = registry_info

            if registry_name == "index.docker.io" and project is None:
                project = "library"

            if not project:
                continue

            query_params.append({
                "registry_id": registry_id,
                "registry_name_and_project": f"{registry_name}/{project}",
            })

        db_connection.execute(
            insert_registry_id_query,
            query_params,
        )


def insert_registry_id_to_images_with_missing_registry_id() -> None:
    """
    If there are image rows with empty registry_id and the image's container registry name exists in the container_registries table,
    Generate new container registry row corresponding to the image's project part.
    """

    db_connection = op.get_bind()
    ContainerRegistryRow = get_container_registry_row_schema()

    get_images_with_blank_registry_id = sa.select(images_table).where(
        images_table.c.registry_id.is_(None)
    )
    images = db_connection.execute(get_images_with_blank_registry_id)

    added_projects = []
    for image in images:
        namespace = (image.name.split("/"))[:2]
        if len(namespace) < 2:
            continue

        registry_name, project = namespace

        registry_info = db_connection.execute(
            sa.select(
                ContainerRegistryRow.url,
                ContainerRegistryRow.registry_name,
                ContainerRegistryRow.type,
                ContainerRegistryRow.ssl_verify,
            ).where(ContainerRegistryRow.registry_name == registry_name)
        ).fetchone()

        if not registry_info:
            continue

        if project in added_projects:
            continue
        added_projects.append(project)

        registry_info_dict: dict[str, Any] = dict(registry_info._mapping)
        registry_info_dict["project"] = project

        registry_id = db_connection.execute(
            sa.insert(ContainerRegistryRow)
            .values(**registry_info_dict)
            .returning(ContainerRegistryRow.id)
        ).scalar_one()

        db_connection.execute(
            sa.update(images_table)
            .values(registry_id=registry_id)
            .where(
                images_table.c.name.startswith(f"{registry_name}/{project}")
                & (images_table.c.registry_id.is_(None))
            )
        )

        logger.info(
            f'Following container registry row auto-generated: "{registry_name}" with the project "{project}".'
        )

    if added_projects:
        logger.info(
            "If credential required for the auto-generated container registry rows, you should fill their credential columns manually."
        )
    else:
        logger.info("No container registry row auto-generated.")


def mark_local_images_with_missing_registry_id() -> None:
    """
    If there are image rows with empty registry_id and the image's container registry name does not exist in the container_registries table,
    Generate a local type container registry record and then, consider and mark the images as local images.
    """

    db_connection = op.get_bind()
    ContainerRegistryRow = get_container_registry_row_schema()

    get_images_with_blank_registry_id = sa.select(images_table).where(
        images_table.c.registry_id.is_(None)
    )
    images_with_missing_registry_id = db_connection.execute(get_images_with_blank_registry_id)

    if not images_with_missing_registry_id:
        return

    local_registry_info = {
        "type": ContainerRegistryType.LOCAL,
        "registry_name": "local",
        # url is not used for local registry.
        # but it is required in the old schema (etcd),
        # so, let's put a dummy value for compatibility purposes.
        "url": "http://localhost",
    }

    local_registry_id = db_connection.execute(
        sa.select(ContainerRegistryRow.id).where(
            ContainerRegistryRow.type == ContainerRegistryType.LOCAL
        )
    ).scalar_one_or_none()

    if not local_registry_id:
        local_registry_id = db_connection.execute(
            sa.insert(ContainerRegistryRow)
            .values(**local_registry_info)
            .returning(ContainerRegistryRow.id)
        ).scalar_one()

    db_connection.execute(
        sa.update(images_table)
        .values(
            registry_id=local_registry_id,
            is_local=True,
        )
        .where(images_table.c.registry_id.is_(None))
    )


def upgrade() -> None:
    metadata = sa.MetaData(naming_convention=convention)
    op.create_table(
        "container_registries",
        metadata,
        IDColumn("id"),
        sa.Column("url", sa.String(length=512), nullable=True, index=True),
        sa.Column("registry_name", sa.String(length=50), nullable=True, index=True),
        sa.Column(
            "type",
            StrEnumType(ContainerRegistryType),
            default=ContainerRegistryType.DOCKER,
            server_default=ContainerRegistryType.DOCKER,
            nullable=False,
            index=True,
        ),
        sa.Column("project", sa.String(length=255), nullable=True, index=True),
        sa.Column("username", sa.String(length=255), nullable=True),
        sa.Column("password", sa.String(), nullable=True),
        sa.Column(
            "ssl_verify", sa.Boolean(), server_default=sa.text("true"), nullable=True, index=True
        ),
        sa.Column(
            "is_global", sa.Boolean(), server_default=sa.text("true"), nullable=True, index=True
        ),
    )

    migrate_data_etcd_to_psql()

    op.add_column(
        "images",
        # registry_id should be non-nullable, but it should be nullable before the migration.
        sa.Column("registry_id", GUID, default=None, nullable=True, index=True),
    )

    insert_registry_id_to_images()
    insert_registry_id_to_images_with_missing_registry_id()
    mark_local_images_with_missing_registry_id()
    op.alter_column("images", "registry_id", nullable=False)

    delete_old_etcd_container_registries()


def downgrade() -> None:
    revert_data_psql_to_etcd()

    op.drop_table("container_registries")
    op.drop_column("images", "registry_id")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, override

from pydantic import BaseModel, Field

from ai.backend.common.bgtask.task.base import (
    BaseBackgroundTaskHandler,
    BaseBackgroundTaskManifest,
    BaseBackgroundTaskResult,
)
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.bgtask.types import ManagerBgtaskName
from ai.backend.manager.services.image.actions.purge_images import PurgeImageAction
from ai.backend.manager.services.image.types import ImageRefData

if TYPE_CHECKING:
    from ai.backend.manager.services.processors import Processors

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class PurgedImageData(BaseModel):
    """Data about a purged image."""

    agent_id: str = Field(description="Agent ID where the image was purged")
    image_name: str = Field(description="Name of the purged image")
    reserved_bytes: int = Field(description="Amount of disk space freed by purging this image")


class PurgeImagesTaskResult(BaseBackgroundTaskResult):
    """
    Result of purge images background task.
    Contains information about purged images and any errors encountered.
    """

    total_reserved_bytes: int = Field(description="Total disk space freed by purging all images")
    purged_images: list[PurgedImageData] = Field(description="List of successfully purged images")
    errors: list[str] = Field(description="List of errors encountered during the purge operation")


class PurgeImageSpec(BaseModel):
    """Specification of a container image to purge."""

    name: str = Field(description="Image name")
    registry: str = Field(description="Registry URL where the image is stored")
    architecture: str = Field(description="Image architecture (e.g., x86_64, aarch64)")


class PurgeAgentSpec(BaseModel):
    """Specification for purging images on a specific agent."""

    agent_id: AgentId = Field(description="Agent ID where images will be purged")
    images: list[PurgeImageSpec] = Field(description="List of images to purge from this agent")


class PurgeImagesManifest(BaseBackgroundTaskManifest):
    """
    Manifest for purging container images from agents.
    """

    keys: list[PurgeAgentSpec] = Field(description="List of agent-specific purge specifications")
    force: bool = Field(description="Force purge even if image is in use")
    noprune: bool = Field(description="Skip pruning dangling images after purge")


class PurgeImagesHandler(BaseBackgroundTaskHandler[PurgeImagesManifest, PurgeImagesTaskResult]):
    """
    Background task handler for purging container images from agents.
    """

    _processors: Processors

    def __init__(self, processors: Processors) -> None:
        self._processors = processors

    @classmethod
    @override
    def name(cls) -> ManagerBgtaskName:
        return ManagerBgtaskName.PURGE_IMAGES

    @classmethod
    @override
    def manifest_type(cls) -> type[PurgeImagesManifest]:
        return PurgeImagesManifest

    @override
    async def execute(self, manifest: PurgeImagesManifest) -> PurgeImagesTaskResult:
        total_reserved_bytes = 0
        purged_images: list[PurgedImageData] = []
        errors: list[str] = []

        for key in manifest.keys:
            for img in key.images:
                result = await self._processors.image.purge_image.wait_for_complete(
                    PurgeImageAction(
                        ImageRefData(
                            name=img.name,
                            registry=img.registry,
                            architecture=img.architecture,
                        ),
                        agent_id=key.agent_id,
                        force=manifest.force,
                        noprune=manifest.noprune,
                    )
                )

                total_reserved_bytes += result.reserved_bytes
                purged_images.append(
                    PurgedImageData(
                        agent_id=key.agent_id,
                        image_name=result.purged_image.name,
                        reserved_bytes=result.reserved_bytes,
                    )
                )

                if result.error is not None:
                    log.error(result.error)
                    errors.append(result.error)

        return PurgeImagesTaskResult(
            total_reserved_bytes=total_reserved_bytes,
            purged_images=purged_images,
            errors=errors,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import ScopeType
from ai.backend.manager.actions.action import SearchActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.types import ScopeData
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class SearchScopesAction(RoleAction):
    """Action to search scopes.

    This action is only available to superadmins.
    Permission check is performed at the API handler level.
    """

    scope_type: ScopeType
    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchScopesActionResult(SearchActionResult[ScopeData]):
    pass



"""GraphQL types for model revisions."""

from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from enum import StrEnum
from pathlib import PurePosixPath
from typing import TYPE_CHECKING, Annotated, Any, Self, cast, override
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID, PageInfo
from strawberry.scalars import JSON

from ai.backend.common.types import ClusterMode as CommonClusterMode
from ai.backend.common.types import MountPermission as CommonMountPermission
from ai.backend.common.types import RuntimeVariant
from ai.backend.manager.api.gql.base import (
    OrderDirection,
    StringFilter,
    to_global_id,
)
from ai.backend.manager.api.gql.common.types import (
    ResourceOptsGQL,
    ResourceOptsInput,
)
from ai.backend.manager.api.gql.fair_share.types.common import ResourceSlotGQL
from ai.backend.manager.api.gql.image_federation import Image
from ai.backend.manager.api.gql.resource_group.federation import ResourceGroup
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.api.gql.vfolder import ExtraVFolderMountConnection, VFolder
from ai.backend.manager.api.gql_legacy.image import ImageNode
from ai.backend.manager.api.gql_legacy.scaling_group import ScalingGroupNode
from ai.backend.manager.api.gql_legacy.vfolder import VirtualFolderNode
from ai.backend.manager.data.deployment.creator import ModelRevisionCreator, VFolderMountsCreator
from ai.backend.manager.data.deployment.types import (
    ClusterConfigData,
    ExecutionSpec,
    ModelMountConfigData,
    ModelRevisionData,
    ModelRevisionOrderField,
    ModelRuntimeConfigData,
    MountInfo,
    ResourceConfigData,
    ResourceSpec,
)
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.deployment.options import RevisionConditions, RevisionOrders

if TYPE_CHECKING:
    from .deployment import ModelDeployment

MountPermission: type[CommonMountPermission] = strawberry.enum(
    CommonMountPermission,
    name="MountPermission",
    description="Added in 25.19.0. This enum represents the permission level for a mounted volume. It can be ro, rw, wd",
)


@strawberry.enum(description="Added in 25.19.0")
class ClusterMode(StrEnum):
    SINGLE_NODE = "SINGLE_NODE"
    MULTI_NODE = "MULTI_NODE"


@strawberry.type(
    name="EnvironmentVariableEntry",
    description="Added in 26.1.0. A single environment variable entry with name and value.",
)
class EnvironmentVariableEntryGQL:
    """A single environment variable entry with name and value."""

    name: str = strawberry.field(
        description="Environment variable name (e.g., CUDA_VISIBLE_DEVICES)."
    )
    value: str = strawberry.field(description="Environment variable value.")


@strawberry.type(
    name="EnvironmentVariables",
    description="Added in 26.1.0. A collection of environment variable entries.",
)
class EnvironmentVariablesGQL:
    """A collection of environment variable entries."""

    entries: list[EnvironmentVariableEntryGQL] = strawberry.field(
        description="List of environment variable entries."
    )


@strawberry.type
class ModelMountConfig:
    """
    Added in 25.19.0.

    Configuration for mounting the model data into the inference container.
    Specifies the virtual folder, mount destination, and model definition path.
    """

    _vfolder_id: strawberry.Private[UUID]
    mount_destination: str = strawberry.field(
        description="Path inside the container where the model is mounted."
    )
    definition_path: str = strawberry.field(
        description="Path to the model definition file within the mounted folder."
    )

    @strawberry.field
    async def vfolder(self, info: Info[StrawberryGQLContext]) -> VFolder:
        vfolder_global_id = to_global_id(
            VirtualFolderNode, self._vfolder_id, is_target_graphene_object=True
        )
        return VFolder(id=ID(vfolder_global_id))

    @classmethod
    def from_dataclass(cls, data: ModelMountConfigData) -> ModelMountConfig | None:
        if data.vfolder_id is None or data.mount_destination is None:
            return None
        return cls(
            _vfolder_id=data.vfolder_id,
            mount_destination=data.mount_destination,
            definition_path=data.definition_path,
        )


@strawberry.type
class ModelRuntimeConfig:
    """
    Added in 25.19.0.

    Runtime configuration for the inference framework.
    Includes the runtime variant, framework-specific configuration,
    and environment variables.
    """

    runtime_variant: str = strawberry.field(
        description="The inference runtime variant (e.g., vllm, triton)."
    )
    inference_runtime_config: JSON | None = strawberry.field(
        description="Framework-specific configuration in JSON format.",
        default=None,
    )
    environ: EnvironmentVariablesGQL | None = strawberry.field(
        description="Environment variables for the service, e.g. CUDA_VISIBLE_DEVICES=0.",
        default=None,
    )

    @classmethod
    def from_dataclass(cls, data: ModelRuntimeConfigData) -> ModelRuntimeConfig:
        environ_gql: EnvironmentVariablesGQL | None = None
        if data.environ is not None:
            environ_gql = EnvironmentVariablesGQL(
                entries=[
                    EnvironmentVariableEntryGQL(name=key, value=value)
                    for key, value in data.environ.items()
                ]
            )
        return cls(
            runtime_variant=data.runtime_variant,
            inference_runtime_config=data.inference_runtime_config,
            environ=environ_gql,
        )


@strawberry.type
class ResourceConfig:
    """
    Added in 25.19.0.

    Compute resource configuration for the deployment.
    Specifies CPU, memory, GPU allocations and additional resource options.
    """

    _resource_group_name: strawberry.Private[str]
    resource_slots: ResourceSlotGQL = strawberry.field(
        description="Added in 26.1.0. Allocated compute resources including CPU, memory, and accelerators."
    )
    resource_opts: ResourceOptsGQL | None = strawberry.field(
        description="Added in 26.1.0. Additional resource options such as shared memory.",
        default=None,
    )

    @strawberry.field
    def resource_group(self) -> ResourceGroup:
        """Resolves the federated ResourceGroup."""
        global_id = to_global_id(
            ScalingGroupNode, self._resource_group_name, is_target_graphene_object=True
        )
        return ResourceGroup(id=ID(global_id))

    @classmethod
    def from_dataclass(cls, data: ResourceConfigData) -> ResourceConfig:
        return cls(
            _resource_group_name=data.resource_group_name,
            resource_slots=ResourceSlotGQL.from_resource_slot(data.resource_slot),
            resource_opts=ResourceOptsGQL.from_mapping(data.resource_opts)
            if data.resource_opts
            else None,
        )


@strawberry.type
class ClusterConfig:
    """
    Added in 25.19.0.

    Cluster configuration for model deployment replicas.
    Defines the clustering mode and number of replicas.
    """

    mode: ClusterMode = strawberry.field(description="The clustering mode (e.g., SINGLE_NODE).")
    size: int = strawberry.field(description="Number of replicas in the cluster.")

    @classmethod
    def from_dataclass(cls, data: ClusterConfigData) -> ClusterConfig:
        return cls(
            mode=ClusterMode(data.mode.name),
            size=data.size,
        )


@strawberry.type
class ModelRevision(Node):
    """
    Added in 25.19.0.

    Represents a versioned configuration snapshot of a model deployment.
    Each revision captures the complete configuration including cluster settings,
    resource allocations, runtime configuration, and model mount settings.

    Revisions enable version control and rollback capabilities for deployments.
    """

    _image_id: strawberry.Private[UUID]
    id: NodeID[str]
    name: str = strawberry.field(description="The name identifier for this revision.")
    cluster_config: ClusterConfig = strawberry.field(
        description="Cluster configuration for replica distribution."
    )
    resource_config: ResourceConfig = strawberry.field(
        description="Compute resource allocation settings."
    )
    model_runtime_config: ModelRuntimeConfig = strawberry.field(
        description="Runtime configuration for the inference framework."
    )
    model_mount_config: ModelMountConfig | None = strawberry.field(
        description="Model data mount configuration."
    )
    extra_mounts: ExtraVFolderMountConnection = strawberry.field(
        description="Additional volume folder mounts."
    )
    created_at: datetime = strawberry.field(description="Timestamp when the revision was created.")

    @strawberry.field
    async def image(self, info: Info[StrawberryGQLContext]) -> Image:
        image_global_id = to_global_id(ImageNode, self._image_id, is_target_graphene_object=True)
        return Image(id=ID(image_global_id))

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.revision_loader.load_many([
            UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: ModelRevisionData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            cluster_config=ClusterConfig.from_dataclass(data.cluster_config),
            resource_config=ResourceConfig.from_dataclass(data.resource_config),
            model_runtime_config=ModelRuntimeConfig.from_dataclass(data.model_runtime_config),
            model_mount_config=ModelMountConfig.from_dataclass(data.model_mount_config),
            extra_mounts=ExtraVFolderMountConnection.from_dataclass(data.extra_vfolder_mounts),
            _image_id=data.image_id,
            created_at=data.created_at,
        )


# Filter and Order Types
@strawberry.input(description="Added in 25.19.0")
class ModelRevisionFilter(GQLFilter):
    name: StringFilter | None = None
    deployment_id: ID | None = None
    ids_in: strawberry.Private[Sequence[UUID] | None] = None

    AND: list[ModelRevisionFilter] | None = None
    OR: list[ModelRevisionFilter] | None = None
    NOT: list[ModelRevisionFilter] | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter.

        Returns a list of QueryCondition callables that can be applied to SQLAlchemy queries.
        """
        field_conditions: list[QueryCondition] = []

        # Apply name filter
        if self.name:
            name_condition = self.name.build_query_condition(
                contains_factory=RevisionConditions.by_name_contains,
                equals_factory=RevisionConditions.by_name_equals,
                starts_with_factory=RevisionConditions.by_name_starts_with,
                ends_with_factory=RevisionConditions.by_name_ends_with,
            )
            if name_condition:
                field_conditions.append(name_condition)

        # Apply deployment_id filter
        if self.deployment_id:
            field_conditions.append(RevisionConditions.by_deployment_id(UUID(self.deployment_id)))

        # Apply ids_in filter
        if self.ids_in:
            field_conditions.append(RevisionConditions.by_ids(self.ids_in))

        # Handle AND logical operator - these are implicitly ANDed with field conditions
        if self.AND:
            for sub_filter in self.AND:
                field_conditions.extend(sub_filter.build_conditions())

        # Handle OR logical operator
        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                field_conditions.append(combine_conditions_or(or_sub_conditions))

        # Handle NOT logical operator
        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                field_conditions.append(negate_conditions(not_sub_conditions))

        return field_conditions


@strawberry.input(description="Added in 25.19.0")
class ModelRevisionOrderBy(GQLOrderBy):
    field: ModelRevisionOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case ModelRevisionOrderField.NAME:
                return RevisionOrders.name(ascending)
            case ModelRevisionOrderField.CREATED_AT:
                return RevisionOrders.created_at(ascending)


# Payload Types
@strawberry.type(description="Added in 25.19.0")
class AddRevisionPayload:
    revision: ModelRevision


@strawberry.type(description="Added in 25.16.0")
class CreateRevisionPayload:
    revision: ModelRevision


@strawberry.input(
    name="ActivateRevisionInput",
    description="Added in 25.19.0. Input for activating a revision to be the current revision.",
)
class ActivateRevisionInputGQL:
    deployment_id: ID
    revision_id: ID


@strawberry.type(
    name="ActivateRevisionPayload",
    description="Added in 25.19.0. Result of activating a revision.",
)
class ActivateRevisionPayloadGQL:
    deployment: Annotated[ModelDeployment, strawberry.lazy(".deployment")]
    previous_revision_id: ID | None
    activated_revision_id: ID


# Input Types
@strawberry.input(description="Added in 25.19.0")
class ClusterConfigInput:
    mode: ClusterMode
    size: int


@strawberry.input(description="Added in 25.19.0")
class ResourceGroupInput:
    name: str


@strawberry.input(
    description=(
        "Added in 26.1.0. A single entry representing one resource type and its allocated quantity."
    )
)
class ResourceSlotEntryInput:
    """Single resource slot entry input with resource type and quantity."""

    resource_type: str = strawberry.field(
        description="Resource type identifier (e.g., 'cpu', 'mem', 'cuda.device')."
    )
    quantity: str = strawberry.field(description="Quantity of the resource as a decimal string.")


@strawberry.input(
    description=("Added in 26.1.0. A collection of compute resource allocations for input.")
)
class ResourceSlotInput:
    """Resource slot input containing multiple resource type entries."""

    entries: list[ResourceSlotEntryInput] = strawberry.field(
        description="List of resource allocations."
    )


@strawberry.input(description="Added in 25.19.0")
class ResourceConfigInput:
    resource_group: ResourceGroupInput
    resource_slots: ResourceSlotInput = strawberry.field(
        description="Added in 26.1.0. Resources allocated for the deployment."
    )
    resource_opts: ResourceOptsInput | None = strawberry.field(
        description="Added in 26.1.0. Additional options for the resources such as shared memory.",
        default=None,
    )


@strawberry.input(description="Added in 25.19.0")
class ImageInput:
    id: ID


@strawberry.input(
    name="EnvironmentVariableEntryInput",
    description="Added in 26.1.0. A single environment variable entry with name and value.",
)
class EnvironmentVariableEntryInputGQL:
    """A single environment variable entry with name and value."""

    name: str = strawberry.field(
        description="Environment variable name (e.g., CUDA_VISIBLE_DEVICES)."
    )
    value: str = strawberry.field(description="Environment variable value.")


@strawberry.input(
    name="EnvironmentVariablesInput",
    description="Added in 26.1.0. A collection of environment variable entries.",
)
class EnvironmentVariablesInputGQL:
    """A collection of environment variable entries."""

    entries: list[EnvironmentVariableEntryInputGQL] = strawberry.field(
        description="List of environment variable entries."
    )


@strawberry.input(description="Added in 25.19.0")
class ModelRuntimeConfigInput:
    runtime_variant: str
    inference_runtime_config: JSON | None = None
    environ: EnvironmentVariablesInputGQL | None = strawberry.field(
        description="Environment variables for the service.",
        default=None,
    )


@strawberry.input(description="Added in 25.19.0")
class ModelMountConfigInput:
    vfolder_id: ID
    mount_destination: str
    definition_path: str


@strawberry.input(description="Added in 25.19.0")
class ExtraVFolderMountInput:
    vfolder_id: ID
    mount_destination: str | None


@strawberry.input(
    description="Added in 25.19.0. Input for creating a revision without attaching to a deployment."
)
class CreateRevisionInput:
    name: str | None = None
    cluster_config: ClusterConfigInput
    resource_config: ResourceConfigInput
    image: ImageInput
    model_runtime_config: ModelRuntimeConfigInput
    model_mount_config: ModelMountConfigInput
    extra_mounts: list[ExtraVFolderMountInput] | None

    def to_model_revision_creator(self) -> ModelRevisionCreator:
        resource_spec = ResourceSpec(
            cluster_mode=CommonClusterMode(self.cluster_config.mode),
            cluster_size=self.cluster_config.size,
            resource_slots=cast(Mapping[str, Any], self.resource_config.resource_slots),
            resource_opts=cast(Mapping[str, Any] | None, self.resource_config.resource_opts),
        )

        extra_mounts = []
        if self.extra_mounts is not None:
            extra_mounts = [
                MountInfo(
                    vfolder_id=UUID(str(extra_mount.vfolder_id)),
                    kernel_path=PurePosixPath(
                        extra_mount.mount_destination
                        if extra_mount.mount_destination is not None
                        else ""
                    ),
                )
                for extra_mount in self.extra_mounts
            ]

        mounts = VFolderMountsCreator(
            model_vfolder_id=UUID(str(self.model_mount_config.vfolder_id)),
            model_definition_path=self.model_mount_config.definition_path,
            model_mount_destination=self.model_mount_config.mount_destination,
            extra_mounts=extra_mounts,
        )

        execution_spec = ExecutionSpec(
            environ={e.name: e.value for e in self.model_runtime_config.environ.entries}
            if self.model_runtime_config.environ
            else None,
            runtime_variant=RuntimeVariant(self.model_runtime_config.runtime_variant),
            inference_runtime_config=cast(
                dict[str, Any] | None, self.model_runtime_config.inference_runtime_config
            ),
        )

        return ModelRevisionCreator(
            image_id=UUID(str(self.image.id)),
            resource_spec=resource_spec,
            mounts=mounts,
            execution=execution_spec,
        )


@strawberry.input(description="Added in 25.19.0")
class AddRevisionInput:
    name: str | None = None
    deployment_id: ID
    cluster_config: ClusterConfigInput
    resource_config: ResourceConfigInput
    image: ImageInput
    model_runtime_config: ModelRuntimeConfigInput
    model_mount_config: ModelMountConfigInput
    extra_mounts: list[ExtraVFolderMountInput] | None

    def to_model_revision_creator(self) -> ModelRevisionCreator:
        """Build ModelRevisionCreator from input fields."""
        resource_spec = ResourceSpec(
            cluster_mode=CommonClusterMode(self.cluster_config.mode),
            cluster_size=self.cluster_config.size,
            resource_slots=cast(Mapping[str, Any], self.resource_config.resource_slots),
            resource_opts=cast(Mapping[str, Any] | None, self.resource_config.resource_opts),
        )

        extra_mounts = []
        if self.extra_mounts is not None:
            extra_mounts = [
                MountInfo(
                    vfolder_id=UUID(str(extra_mount.vfolder_id)),
                    kernel_path=PurePosixPath(
                        extra_mount.mount_destination
                        if extra_mount.mount_destination is not None
                        else ""
                    ),
                )
                for extra_mount in self.extra_mounts
            ]

        mounts = VFolderMountsCreator(
            model_vfolder_id=UUID(str(self.model_mount_config.vfolder_id)),
            model_definition_path=self.model_mount_config.definition_path,
            model_mount_destination=self.model_mount_config.mount_destination,
            extra_mounts=extra_mounts,
        )

        execution_spec = ExecutionSpec(
            environ={e.name: e.value for e in self.model_runtime_config.environ.entries}
            if self.model_runtime_config.environ
            else None,
            runtime_variant=RuntimeVariant(self.model_runtime_config.runtime_variant),
            inference_runtime_config=cast(
                dict[str, Any] | None, self.model_runtime_config.inference_runtime_config
            ),
        )

        return ModelRevisionCreator(
            image_id=UUID(str(self.image.id)),
            resource_spec=resource_spec,
            mounts=mounts,
            execution=execution_spec,
        )


ModelRevisionEdge = Edge[ModelRevision]


@strawberry.type(description="Added in 25.19.0")
class ModelRevisionConnection(Connection[ModelRevision]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count

    @classmethod
    def from_dataclass(cls, revisions_data: list[ModelRevisionData]) -> ModelRevisionConnection:
        nodes = [ModelRevision.from_dataclass(data) for data in revisions_data]
        edges = [ModelRevisionEdge(node=node, cursor=str(node.id)) for node in nodes]

        page_info = PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        )

        return cls(count=len(nodes), edges=edges, page_info=page_info)



from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.types import ModelDeploymentData
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class GetDeploymentByIdAction(DeploymentBaseAction):
    deployment_id: UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.deployment_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetDeploymentByIdActionResult(BaseActionResult):
    data: ModelDeploymentData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



"""add routing revision and traffic_status columns

Revision ID: 6f8a4c0d2e3g
Revises: 5e7a3b9c1d2f
Create Date: 2025-12-17

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "6f8a4c0d2e3g"
down_revision = "5e7a3b9c1d2f"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # 1. Create RouteTrafficStatus enum type (active/inactive only)
    op.execute("CREATE TYPE routetrafficstatus AS ENUM ('active', 'inactive')")

    # 2. Add revision column (nullable, no FK constraint)
    op.add_column(
        "routings",
        sa.Column("revision", postgresql.UUID(as_uuid=True), nullable=True),
    )

    # 3. Add traffic_status column with default 'active'
    op.add_column(
        "routings",
        sa.Column(
            "traffic_status",
            postgresql.ENUM(
                "active",
                "inactive",
                name="routetrafficstatus",
                create_type=False,
            ),
            nullable=False,
            server_default=sa.text("'active'"),
        ),
    )

    # 4. Set revision for existing routes from endpoint.current_revision
    op.execute("""
        UPDATE routings r
        SET revision = e.current_revision
        FROM endpoints e
        WHERE r.endpoint = e.id
          AND e.current_revision IS NOT NULL
    """)


def downgrade() -> None:
    op.drop_column("routings", "traffic_status")
    op.drop_column("routings", "revision")
    op.execute("DROP TYPE routetrafficstatus")



from ai.backend.manager.models.hasher.types import PasswordColumn, PasswordInfo

from .row import (
    ACTIVE_USER_STATUSES,
    INACTIVE_USER_STATUSES,
    PasswordHashAlgorithm,
    UserRole,
    UserRow,
    UserStatus,
    by_user_email,
    by_user_uuid,
    by_username,
    check_credential,
    check_credential_with_migration,
    compare_to_hashed_password,
    users,
)

__all__ = (
    "ACTIVE_USER_STATUSES",
    "INACTIVE_USER_STATUSES",
    "PasswordColumn",
    "PasswordHashAlgorithm",
    "PasswordInfo",
    "UserRole",
    "UserRow",
    "UserStatus",
    "by_user_email",
    "by_user_uuid",
    "by_username",
    "check_credential",
    "check_credential_with_migration",
    "compare_to_hashed_password",
    "users",
)



"""Validator for group resource limits."""

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.sokovan.data import SessionWorkload, SystemSnapshot

from .exceptions import GroupResourceQuotaExceeded
from .validator import ValidatorRule


class GroupResourceLimitValidator(ValidatorRule):
    """
    Check if a session would exceed the group's resource quota.
    This corresponds to check_group_resource_limit predicate.
    """

    def name(self) -> str:
        """Return the validator name for predicates."""
        return "GroupResourceLimitValidator"

    def success_message(self) -> str:
        """Return a message describing successful validation."""
        return "Group has sufficient resource quota for the requested session"

    def validate(self, snapshot: SystemSnapshot, workload: SessionWorkload) -> None:
        # Get the group's resource limit
        group_limit = snapshot.resource_policy.group_limits.get(workload.group_id)
        if not group_limit:
            # If no limit is defined, we can't validate - let it pass
            return

        # Get current group occupancy
        group_occupied = snapshot.resource_occupancy.by_group.get(workload.group_id, ResourceSlot())

        # Check if adding this workload would exceed the limit
        total_after = group_occupied + workload.requested_slots
        if not (total_after <= group_limit):
            # Format the limit for human-readable output
            if group_limit and any(v for v in group_limit.values()):
                limit_str = " ".join(f"{k}={v}" for k, v in group_limit.items() if v)
                exceeded_msg = f"limit: {limit_str}, current: {group_occupied}, requested: {workload.requested_slots}"
            else:
                exceeded_msg = f"No resource limits defined. current: {group_occupied}, requested: {workload.requested_slots}"
            raise GroupResourceQuotaExceeded(
                f"Your group resource quota is exceeded. ({exceeded_msg})"
            )



from __future__ import annotations

import uuid
from collections.abc import Mapping

from ai.backend.common.data.permission.types import GLOBAL_SCOPE_ID
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.permission.entity import EntityListResult
from ai.backend.manager.data.permission.id import ObjectId, ScopeId
from ai.backend.manager.data.permission.object_permission import (
    ObjectPermissionData,
    ObjectPermissionListResult,
)
from ai.backend.manager.data.permission.permission import (
    PermissionData,
    PermissionListResult,
)
from ai.backend.manager.data.permission.role import (
    AssignedUserListResult,
    BatchEntityPermissionCheckInput,
    RoleData,
    RoleDetailData,
    RoleListResult,
    RolePermissionsUpdateInput,
    ScopePermissionCheckInput,
    SingleEntityPermissionCheckInput,
    UserRoleAssignmentData,
    UserRoleAssignmentInput,
    UserRoleRevocationData,
    UserRoleRevocationInput,
)
from ai.backend.manager.data.permission.types import (
    ScopeData,
    ScopeListResult,
    ScopeType,
)
from ai.backend.manager.models.rbac_models.permission.object_permission import ObjectPermissionRow
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow
from ai.backend.manager.models.rbac_models.role import RoleRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.purger import Purger
from ai.backend.manager.repositories.base.querier import BatchQuerier
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.permission_controller.types import (
    ObjectPermissionSearchScope,
    PermissionSearchScope,
)

from .db_source.db_source import CreateRoleInput, PermissionDBSource

permission_controller_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.REPOSITORY, layer=LayerType.PERMISSION_CONTROLLER_REPOSITORY
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class PermissionControllerRepository:
    _db_source: PermissionDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = PermissionDBSource(db)

    @permission_controller_repository_resilience.apply()
    async def create_role(self, input_data: CreateRoleInput) -> RoleData:
        """
        Create a new role in the database.

        Returns the created role data.
        """
        role_row = await self._db_source.create_role(input_data)
        return role_row.to_data()

    @permission_controller_repository_resilience.apply()
    async def create_permission(
        self,
        creator: Creator[PermissionRow],
    ) -> PermissionData:
        """
        Create a new permission in the database.

        Returns the created permission data.
        """
        row = await self._db_source.create_permission(creator)
        return row.to_data()

    @permission_controller_repository_resilience.apply()
    async def delete_permission(
        self,
        purger: Purger[PermissionRow],
    ) -> PermissionData:
        """
        Delete a permission from the database.

        Returns the deleted permission data.

        Raises:
            ObjectNotFound: If permission does not exist.
        """
        row = await self._db_source.delete_permission(purger)
        return row.to_data()

    @permission_controller_repository_resilience.apply()
    async def create_object_permission(
        self,
        creator: Creator[ObjectPermissionRow],
    ) -> ObjectPermissionData:
        """
        Create a new object permission in the database.

        Returns the created object permission data.
        """
        row = await self._db_source.create_object_permission(creator)
        return row.to_data()

    @permission_controller_repository_resilience.apply()
    async def delete_object_permission(
        self,
        purger: Purger[ObjectPermissionRow],
    ) -> ObjectPermissionData | None:
        """
        Delete an object permission from the database.

        Returns the deleted object permission data, or None if not found.
        """
        row = await self._db_source.delete_object_permission(purger)
        return row.to_data() if row else None

    @permission_controller_repository_resilience.apply()
    async def update_role(self, updater: Updater[RoleRow]) -> RoleData:
        result = await self._db_source.update_role(updater)
        return result.to_data()

    @permission_controller_repository_resilience.apply()
    async def update_role_permissions(
        self, input_data: RolePermissionsUpdateInput
    ) -> RoleDetailData:
        """Update role permissions using batch update."""
        result = await self._db_source.update_role_permissions(input_data=input_data)
        return result.to_detail_data_without_users()

    @permission_controller_repository_resilience.apply()
    async def delete_role(self, updater: Updater[RoleRow]) -> RoleData:
        result = await self._db_source.delete_role(updater)
        return result.to_data()

    @permission_controller_repository_resilience.apply()
    async def purge_role(self, purger: Purger[RoleRow]) -> RoleData:
        result = await self._db_source.purge_role(purger)
        return result.to_data()

    @permission_controller_repository_resilience.apply()
    async def assign_role(self, data: UserRoleAssignmentInput) -> UserRoleAssignmentData:
        result = await self._db_source.assign_role(data)
        return result.to_data()

    @permission_controller_repository_resilience.apply()
    async def revoke_role(self, data: UserRoleRevocationInput) -> UserRoleRevocationData:
        user_role_id = await self._db_source.revoke_role(data)
        return UserRoleRevocationData(
            user_role_id=user_role_id, user_id=data.user_id, role_id=data.role_id
        )

    @permission_controller_repository_resilience.apply()
    async def get_role(self, role_id: uuid.UUID) -> RoleData | None:
        result = await self._db_source.get_role(role_id)
        return result.to_data() if result else None

    @permission_controller_repository_resilience.apply()
    async def check_permission_of_entity(self, data: SingleEntityPermissionCheckInput) -> bool:
        target_object_id = data.target_object_id
        roles = await self._db_source.get_user_roles(data.user_id)
        for role in roles:
            for object_perm in role.object_permission_rows:
                if object_perm.operation != data.operation:
                    continue
                if object_perm.object_id() == target_object_id:
                    return True
        return False

    @permission_controller_repository_resilience.apply()
    async def check_permission_in_scope(self, data: ScopePermissionCheckInput) -> bool:
        return await self._db_source.check_scope_permission_exist(
            data.user_id, data.target_scope_id, data.operation
        )

    @permission_controller_repository_resilience.apply()
    async def check_permission_of_entities(
        self,
        data: BatchEntityPermissionCheckInput,
    ) -> Mapping[ObjectId, bool]:
        """
        Check if the user has the requested operation permission on the given entity IDs.
        Returns a mapping of entity ID to a boolean indicating permission.
        """
        return await self._db_source.check_batch_object_permission_exist(
            data.user_id, data.target_object_ids, data.operation
        )

    @permission_controller_repository_resilience.apply()
    async def search_roles(
        self,
        querier: BatchQuerier,
    ) -> RoleListResult:
        """Searches roles with pagination and filtering."""
        return await self._db_source.search_roles(querier=querier)

    @permission_controller_repository_resilience.apply()
    async def search_permissions(
        self,
        querier: BatchQuerier,
        scope: PermissionSearchScope | None = None,
    ) -> PermissionListResult:
        """Searches permissions with pagination and filtering."""
        return await self._db_source.search_permissions(querier=querier, scope=scope)

    @permission_controller_repository_resilience.apply()
    async def search_object_permissions(
        self,
        querier: BatchQuerier,
        scope: ObjectPermissionSearchScope | None = None,
    ) -> ObjectPermissionListResult:
        """Searches object permissions with pagination and filtering."""
        return await self._db_source.search_object_permissions(querier=querier, scope=scope)

    @permission_controller_repository_resilience.apply()
    async def get_role_with_permissions(self, role_id: uuid.UUID) -> RoleDetailData:
        """Get role with all permission details (without users)."""
        result = await self._db_source.get_role_with_permissions(role_id)
        return result.to_detail_data_without_users()

    @permission_controller_repository_resilience.apply()
    async def search_users_assigned_to_role(
        self,
        querier: BatchQuerier,
    ) -> AssignedUserListResult:
        """Searches users assigned to a specific role with pagination and filtering."""
        return await self._db_source.search_users_assigned_to_role(
            querier=querier,
        )

    def _get_global_scope(self) -> ScopeListResult:
        """Get the global scope as a static result."""
        return ScopeListResult(
            items=[
                ScopeData(
                    id=ScopeId(scope_type=ScopeType.GLOBAL, scope_id=GLOBAL_SCOPE_ID),
                    name=GLOBAL_SCOPE_ID,
                )
            ],
            total_count=1,
            has_next_page=False,
            has_previous_page=False,
        )

    @permission_controller_repository_resilience.apply()
    async def search_scopes(
        self,
        scope_type: ScopeType,
        querier: BatchQuerier,
    ) -> ScopeListResult:
        """Search scopes based on scope type.

        Args:
            scope_type: The type of scope to search.
            querier: BatchQuerier with conditions, orders, and pagination.

        Returns:
            ScopeListResult with matching scopes.
        """
        match scope_type:
            case ScopeType.GLOBAL:
                return self._get_global_scope()
            case ScopeType.DOMAIN:
                return await self._db_source.search_domain_scopes(querier)
            case ScopeType.PROJECT:
                return await self._db_source.search_project_scopes(querier)
            case ScopeType.USER:
                return await self._db_source.search_user_scopes(querier)
            case _:
                raise NotImplementedError(
                    "This function will be deprecated and new repository functions will be implemented for each scope"
                )

    @permission_controller_repository_resilience.apply()
    async def search_entities(
        self,
        querier: BatchQuerier,
    ) -> EntityListResult:
        """Search entities within a scope.

        Args:
            querier: BatchQuerier with scope conditions and pagination settings.

        Returns:
            EntityListResult with matching entities.
        """
        return await self._db_source.search_entities_in_scope(querier)



"""Add vfs_storages table

Revision ID: de1032a11cca
Revises: fa3dd7f77c19
Create Date: 2025-10-13 05:45:41.699271

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "de1032a11cca"
down_revision = "fa3dd7f77c19"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "vfs_storages",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("host", sa.String(), nullable=False),
        sa.Column("base_path", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_vfs_storages")),
    )
    op.create_index(op.f("ix_vfs_storages_name"), "vfs_storages", ["name"], unique=True)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_vfs_storages_name"), table_name="vfs_storages")
    op.drop_table("vfs_storages")
    # ### end Alembic commands ###



import uuid
from collections.abc import Mapping
from dataclasses import dataclass
from datetime import timedelta
from typing import Any, override

import yarl

from ai.backend.common.types import AccessKey, ClusterMode, SessionTypes
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.api.utils import Undefined
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.session.base import SessionAction


# TODO: Idea: Refactor this type using pydantic and utilize as API model
# TODO: Remove Undefined before passing to Service layer
@dataclass
class CreateFromTemplateActionParams:
    template_id: uuid.UUID
    session_name: str | Undefined
    image: str | Undefined
    architecture: str | Undefined
    session_type: SessionTypes | Undefined
    group_name: str | Undefined
    domain_name: str | Undefined
    cluster_size: int
    cluster_mode: ClusterMode
    config: dict[str, Any]
    tag: str | Undefined
    priority: int
    owner_access_key: AccessKey | Undefined
    enqueue_only: bool
    max_wait_seconds: int
    starts_at: str | None
    reuse_if_exists: bool
    startup_command: str | None
    batch_timeout: timedelta | None
    bootstrap_script: str | None | Undefined
    dependencies: list[uuid.UUID] | None
    callback_url: yarl.URL | None


@dataclass
class CreateFromTemplateAction(SessionAction):
    params: CreateFromTemplateActionParams
    user_id: uuid.UUID
    user_role: UserRole
    sudo_session_enabled: bool
    requester_access_key: AccessKey
    keypair_resource_policy: dict[str, Any] | None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateFromTemplateActionResult(BaseActionResult):
    session_id: uuid.UUID

    # TODO: Add proper type
    result: Mapping[str, Any]

    @override
    def entity_id(self) -> str | None:
        return str(self.session_id)



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class SearchImagesAction(ImageAction):
    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchImagesActionResult(BaseActionResult):
    data: list[ImageData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""Deployment policy action definitions."""

from .get_deployment_policy import (
    GetDeploymentPolicyAction,
    GetDeploymentPolicyActionResult,
)

__all__ = [
    "GetDeploymentPolicyAction",
    "GetDeploymentPolicyActionResult",
]



from ai.backend.manager.data.kernel.types import KernelStatus

from .row import (
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    DEAD_KERNEL_STATUSES,
    DEFAULT_KERNEL_ORDERING,
    KERNEL_STATUS_TRANSITION_MAP,
    LIVE_STATUS,
    RESOURCE_USAGE_KERNEL_STATUSES,
    USER_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    KernelRow,
    by_kernel_ids,
    get_user_email,
    handle_kernel_exception,
    kernels,
    recalc_concurrency_used,
)

__all__ = (
    "AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES",
    "DEAD_KERNEL_STATUSES",
    "DEFAULT_KERNEL_ORDERING",
    "KERNEL_STATUS_TRANSITION_MAP",
    "LIVE_STATUS",
    "RESOURCE_USAGE_KERNEL_STATUSES",
    "USER_RESOURCE_OCCUPYING_KERNEL_STATUSES",
    "KernelRow",
    "KernelStatus",
    "by_kernel_ids",
    "get_user_email",
    "handle_kernel_exception",
    "kernels",
    "recalc_concurrency_used",
)



"""sync_session_occupying_slots_to_sibling_kernels

Revision ID: 679e5721e94d
Revises: f56a82d0ac9f
Create Date: 2024-04-01 17:34:33.480996

"""

import textwrap
from typing import Any, cast

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.orm import Session, load_only, registry, relationship, selectinload
from sqlalchemy.sql import text

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.models.base import GUID, IDColumn, ResourceSlotColumn, convention

# revision identifiers, used by Alembic.
revision = "679e5721e94d"
down_revision = "f56a82d0ac9f"
branch_labels = None
depends_on = None

metadata = sa.MetaData(naming_convention=convention)
mapper_registry = registry(metadata=metadata)
Base: Any = mapper_registry.generate_base()

PAGE_SIZE = 100


class SessionRow(Base):  # type: ignore[misc]
    __tablename__ = "sessions"
    __table_args__ = {"extend_existing": True}

    id = IDColumn()
    cluster_size = sa.Column("cluster_size", sa.Integer, nullable=False, default=1)
    starts_at = sa.Column("starts_at", sa.DateTime(timezone=True), nullable=True, default=sa.null())
    status_history = sa.Column("status_history", pgsql.JSONB(), nullable=True, default=sa.null())
    occupying_slots: sa.Column[ResourceSlot] = sa.Column(
        "occupying_slots", ResourceSlotColumn(), nullable=False
    )

    kernels = relationship("KernelRow")


class KernelRow(Base):  # type: ignore[misc]
    __tablename__ = "kernels"
    __table_args__ = {"extend_existing": True}

    id = IDColumn()
    session_id = sa.Column(
        "session_id",
        GUID,
        sa.ForeignKey("sessions.id"),
        unique=False,
        index=True,
        nullable=False,
    )
    occupied_slots: sa.Column[ResourceSlot] = sa.Column(
        "occupied_slots", ResourceSlotColumn(), nullable=False
    )


def _sync_single_kernel_cluster_session() -> None:
    conn = op.get_bind()
    sync_stmt = textwrap.dedent(
        """
        UPDATE sessions
        SET occupying_slots = kernels.occupied_slots
        FROM kernels
        WHERE sessions.id = kernels.session_id
        AND sessions.cluster_size = 1;
        """
    )
    conn.execute(text(sync_stmt))


def _sync_multi_kernel_cluster_session() -> None:
    db_sess = Session(op.get_bind())

    while True:
        select_stmt = (
            sa.select(SessionRow)
            .where(
                (SessionRow.cluster_size != 1)
                & (SessionRow.occupying_slots == {})
                & (SessionRow.status_history.op("?")("RUNNING"))
            )
            .limit(PAGE_SIZE)
            .options(selectinload(SessionRow.kernels).options(load_only("occupied_slots")))  # type: ignore[arg-type]
        )
        session_list = cast(list[SessionRow], db_sess.scalars(select_stmt).all())
        if not session_list:
            return

        update_stmt = (
            sa.update(SessionRow)
            .where(SessionRow.id == sa.bindparam("session_id"))
            .values(occupying_slots=sa.bindparam("occupying_slots"))
        )
        data = []
        for session in session_list:
            occupying_slots = sum([k.occupied_slots for k in session.kernels], start=ResourceSlot())
            data.append({"session_id": session.id, "occupying_slots": occupying_slots})
        db_sess.execute(update_stmt, data)


def upgrade() -> None:
    _sync_single_kernel_cluster_session()
    _sync_multi_kernel_cluster_session()


def downgrade() -> None:
    pass



"""
Factory for creating password hasher instances.
"""

from __future__ import annotations

import logging

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.auth.hash import PasswordHashAlgorithm

from .base import PasswordHasher
from .bcrypt import BcryptHasher
from .pbkdf2 import PBKDF2_SHA3_256Hasher, PBKDF2_SHA256Hasher
from .sha import SHA3_256Hasher, SHA256Hasher

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class PasswordHasherFactory:
    """Factory for creating password hasher instances."""

    _hashers: dict[PasswordHashAlgorithm, PasswordHasher] = {
        PasswordHashAlgorithm.BCRYPT: BcryptHasher(),
        PasswordHashAlgorithm.SHA256: SHA256Hasher(),
        PasswordHashAlgorithm.SHA3_256: SHA3_256Hasher(),
        PasswordHashAlgorithm.PBKDF2_SHA256: PBKDF2_SHA256Hasher(),
        PasswordHashAlgorithm.PBKDF2_SHA3_256: PBKDF2_SHA3_256Hasher(),
    }

    @classmethod
    def get_hasher(cls, algorithm: PasswordHashAlgorithm) -> PasswordHasher:
        """Get a hasher instance for the specified algorithm."""
        hasher = cls._hashers.get(algorithm)
        if hasher is None:
            raise ValueError(f"Unsupported password hash algorithm: {algorithm}")
        return hasher



"""
Background task-related exceptions.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class InvalidBgtaskId(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-bgtask-id"
    error_title = "Invalid background task ID format."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BGTASK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class DeleteReservoirRegistryAction(ArtifactRegistryAction):
    reservoir_id: uuid.UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.reservoir_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeleteReservoirActionResult(BaseActionResult):
    deleted_reservoir_id: uuid.UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.deleted_reservoir_id)



from .types import (
    NotificationChannelData,
    NotificationChannelListResult,
    NotificationChannelModifier,
    NotificationRuleData,
    NotificationRuleListResult,
    NotificationRuleModifier,
)

__all__ = (
    "NotificationChannelData",
    "NotificationChannelListResult",
    "NotificationChannelModifier",
    "NotificationRuleData",
    "NotificationRuleListResult",
    "NotificationRuleModifier",
)



"""GraphQL types for kernel management."""

from __future__ import annotations

from .types import (
    KernelV2ClusterInfoGQL,
    KernelV2ConnectionGQL,
    KernelV2EdgeGQL,
    KernelV2FilterGQL,
    KernelV2GQL,
    KernelV2LifecycleInfoGQL,
    KernelV2NetworkInfoGQL,
    KernelV2OrderByGQL,
    KernelV2OrderFieldGQL,
    KernelV2ResourceInfoGQL,
    KernelV2SessionInfoGQL,
    KernelV2StatusFilterGQL,
    KernelV2StatusGQL,
    KernelV2UserInfoGQL,
    ResourceAllocationGQL,
)

__all__ = [
    "KernelV2ConnectionGQL",
    "KernelV2ClusterInfoGQL",
    "KernelV2EdgeGQL",
    "KernelV2FilterGQL",
    "KernelV2LifecycleInfoGQL",
    "KernelV2NetworkInfoGQL",
    "KernelV2OrderByGQL",
    "KernelV2OrderFieldGQL",
    "KernelV2ResourceInfoGQL",
    "KernelV2SessionInfoGQL",
    "KernelV2StatusFilterGQL",
    "KernelV2StatusGQL",
    "KernelV2UserInfoGQL",
    "KernelV2GQL",
    "ResourceAllocationGQL",
]



"""Handler for checking session preconditions."""

from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING

from ai.backend.common.types import AccessKey
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus, StatusTransitions, TransitionStatus
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.scheduler import SchedulerRepository
from ai.backend.manager.sokovan.data import SessionWithKernels
from ai.backend.manager.sokovan.scheduler.handlers.base import SessionLifecycleHandler
from ai.backend.manager.sokovan.scheduler.results import (
    SessionExecutionResult,
    SessionTransitionInfo,
)

if TYPE_CHECKING:
    from ai.backend.manager.sokovan.scheduler.launcher.launcher import SessionLauncher

log = BraceStyleAdapter(logging.getLogger(__name__))


class CheckPreconditionLifecycleHandler(SessionLifecycleHandler):
    """Handler for checking session preconditions and triggering image pulling.

    Following the DeploymentCoordinator pattern:
    - Coordinator queries sessions with SCHEDULED status (provides HandlerSessionData)
    - Handler queries additional data (SessionDataForPull + ImageConfigData) via Repository
    - Handler triggers image pulling on agents via Launcher
    - Coordinator updates sessions to PREPARING status and broadcasts events
    """

    def __init__(
        self,
        launcher: SessionLauncher,
        repository: SchedulerRepository,
    ) -> None:
        self._launcher = launcher
        self._repository = repository

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "check-precondition"

    @classmethod
    def target_statuses(cls) -> list[SessionStatus]:
        """Sessions in SCHEDULED state."""
        return [SessionStatus.SCHEDULED]

    @classmethod
    def target_kernel_statuses(cls) -> list[KernelStatus] | None:
        """Include sessions where kernels are in SCHEDULED status."""
        return [KernelStatus.SCHEDULED]

    @classmethod
    def status_transitions(cls) -> StatusTransitions:
        """Define state transitions for check precondition handler (BEP-1030).

        - success: Session/kernel → PREPARING
        - need_retry: None (stays SCHEDULED)
        - expired: Session/kernel → PENDING (re-scheduling after timeout)
        - give_up: None (image pulling is time-based, only timeout applies)
        """
        return StatusTransitions(
            success=TransitionStatus(
                session=SessionStatus.PREPARING,
                kernel=KernelStatus.PREPARING,
            ),
            need_retry=None,
            expired=TransitionStatus(
                session=SessionStatus.PENDING,
                kernel=KernelStatus.PENDING,
            ),
            give_up=None,
        )

    @property
    def lock_id(self) -> LockID | None:
        """Lock for operations targeting SCHEDULED sessions transitioning to PREPARING."""
        return LockID.LOCKID_SOKOVAN_TARGET_PREPARING

    async def execute(
        self,
        _scaling_group: str,
        sessions: Sequence[SessionWithKernels],
    ) -> SessionExecutionResult:
        """Trigger image pulling for SCHEDULED sessions.

        The coordinator provides SessionWithKernels data.
        This handler:
        1. Extracts session IDs from SessionWithKernels
        2. Queries Repository for additional data (SessionDataForPull + ImageConfigData)
        3. Triggers image pulling via Launcher
        """
        result = SessionExecutionResult()

        if not sessions:
            return result

        # Extract session IDs from SessionWithKernels
        session_ids = [s.session_info.identity.id for s in sessions]

        # Query Repository for additional data needed by Launcher
        sessions_for_pull_data = await self._repository.get_sessions_for_pull_by_ids(session_ids)

        # Trigger image pulling via Launcher with the full data
        # Note: RecorderContext is handled inside Launcher
        await self._launcher.trigger_image_pulling(
            sessions_for_pull_data.sessions,
            sessions_for_pull_data.image_configs,
        )

        # Mark all sessions as success for status transition
        for session in sessions:
            session_info = session.session_info
            result.successes.append(
                SessionTransitionInfo(
                    session_id=session_info.identity.id,
                    from_status=session_info.lifecycle.status,
                    reason="passed-preconditions",
                    creation_id=session_info.identity.creation_id,
                    access_key=AccessKey(session_info.metadata.access_key),
                )
            )

        return result



"""Fair share aggregator for computing kernel resource usage.

This module provides the FairShareAggregator that performs pure computation:
1. Prepares kernel usage records (resource-seconds) in 5-minute slices
2. Aggregates usage deltas by user/project/domain for bucket updates
3. Calculates scheduling ranks for fair share sequencing (future)

The aggregator is stateless and does not interact with databases directly.
Repository operations are handled by FairShareObserver.
"""

from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass, field
from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import TYPE_CHECKING
from uuid import UUID

from ai.backend.common.types import ResourceSlot
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.repositories.resource_usage_history import (
    KernelUsageRecordCreatorSpec,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

# Observation slice duration (5 minutes)
SLICE_DURATION_SECONDS = 300

if TYPE_CHECKING:
    from ai.backend.manager.data.kernel.types import KernelInfo


# =============================================================================
# Bucket Key Types
# =============================================================================


@dataclass(frozen=True)
class UserUsageBucketKey:
    """Key for user usage bucket aggregation.

    Uniquely identifies a user's usage bucket within a resource group and time period.
    """

    user_uuid: UUID
    project_id: UUID
    domain_name: str
    resource_group: str
    period_date: date


@dataclass(frozen=True)
class ProjectUsageBucketKey:
    """Key for project usage bucket aggregation.

    Uniquely identifies a project's usage bucket within a resource group and time period.
    """

    project_id: UUID
    domain_name: str
    resource_group: str
    period_date: date


@dataclass(frozen=True)
class DomainUsageBucketKey:
    """Key for domain usage bucket aggregation.

    Uniquely identifies a domain's usage bucket within a resource group and time period.
    """

    domain_name: str
    resource_group: str
    period_date: date


# =============================================================================
# Result Types
# =============================================================================


@dataclass
class KernelUsagePreparationResult:
    """Result of preparing kernel usage records.

    Attributes:
        specs: List of kernel usage record specs for bulk creation
        kernel_observation_times: Mapping of kernel_id to last_observed_at timestamp
        observed_count: Number of kernels with generated specs
    """

    specs: list[KernelUsageRecordCreatorSpec] = field(default_factory=list)
    kernel_observation_times: dict[UUID, datetime] = field(default_factory=dict)
    observed_count: int = 0


@dataclass
class UsageBucketAggregationResult:
    """Result of aggregating kernel usage into hourly buckets.

    Attributes:
        user_usage_deltas: Aggregated usage deltas by user for bucket updates
        project_usage_deltas: Aggregated usage deltas by project for bucket updates
        domain_usage_deltas: Aggregated usage deltas by domain for bucket updates
    """

    user_usage_deltas: dict[UserUsageBucketKey, ResourceSlot] = field(default_factory=dict)
    project_usage_deltas: dict[ProjectUsageBucketKey, ResourceSlot] = field(default_factory=dict)
    domain_usage_deltas: dict[DomainUsageBucketKey, ResourceSlot] = field(default_factory=dict)


class FairShareAggregator:
    """Computes kernel usage for fair share scheduling.

    This class performs pure computation without database interactions.
    It prepares kernel usage records aligned to 5-minute clock boundaries.

    Partial slices are only allowed at:
    - Start: When starts_at is not on a boundary (first observation only)
    - End: When the kernel terminates at a non-boundary time

    For RUNNING kernels, only complete slices up to the last boundary are generated.
    """

    def prepare_kernel_usage_records(
        self,
        kernels: Sequence[KernelInfo],
        scaling_group: str,
        now: datetime,
    ) -> KernelUsagePreparationResult:
        """Prepare kernel usage records for bulk creation.

        For each kernel:
        1. Determine period: last_observed_at (or starts_at) to end_time
           - RUNNING kernels: end_time = floor(now) to last boundary
           - TERMINATED kernels: end_time = terminated_at
        2. Split period into 5-minute slices aligned to clock boundaries
        3. Create usage record specs for each slice

        Args:
            kernels: Kernels to process
            scaling_group: The scaling group name
            now: Current time from DB

        Returns:
            KernelUsagePreparationResult containing specs and metadata
        """
        result = KernelUsagePreparationResult()

        for kernel in kernels:
            kernel_specs, observation_end = self._prepare_kernel_usage_specs(
                kernel, scaling_group, now
            )
            if kernel_specs:
                result.specs.extend(kernel_specs)
                result.kernel_observation_times[UUID(str(kernel.id))] = observation_end
                result.observed_count += 1

        return result

    def aggregate_kernel_usage_to_buckets(
        self,
        specs: Sequence[KernelUsageRecordCreatorSpec],
    ) -> UsageBucketAggregationResult:
        """Aggregate kernel usage specs into daily bucket deltas.

        Splits each spec's resource usage across day boundaries and aggregates
        by user/project/domain. Buckets are aligned to day boundaries (midnight).

        For example, a spec covering 23:57-00:03 will be split:
        - 23:57-00:00 (3 minutes) -> day 1 bucket
        - 00:00-00:03 (3 minutes) -> day 2 bucket

        Args:
            specs: Kernel usage record specs to aggregate

        Returns:
            UsageBucketAggregationResult with deltas for each bucket
        """
        user_deltas: dict[UserUsageBucketKey, ResourceSlot] = defaultdict(ResourceSlot)
        project_deltas: dict[ProjectUsageBucketKey, ResourceSlot] = defaultdict(ResourceSlot)
        domain_deltas: dict[DomainUsageBucketKey, ResourceSlot] = defaultdict(ResourceSlot)

        for spec in specs:
            # Split spec across day boundaries and aggregate
            daily_splits = self._split_spec_by_day(spec)

            for period_date, resource_usage in daily_splits:
                self._add_to_bucket_deltas(
                    spec=spec,
                    period_date=period_date,
                    resource_usage=resource_usage,
                    user_deltas=user_deltas,
                    project_deltas=project_deltas,
                    domain_deltas=domain_deltas,
                )

        return UsageBucketAggregationResult(
            user_usage_deltas=dict(user_deltas),
            project_usage_deltas=dict(project_deltas),
            domain_usage_deltas=dict(domain_deltas),
        )

    def _split_spec_by_day(
        self,
        spec: KernelUsageRecordCreatorSpec,
    ) -> list[tuple[date, ResourceSlot]]:
        """Split a spec's resource usage across day boundaries.

        Most 5-minute specs will fit within a single day, but specs crossing
        midnight (e.g., 23:57-00:02) need to be split.

        Args:
            spec: Kernel usage record spec to split

        Returns:
            List of (period_date, resource_usage) tuples
        """
        result: list[tuple[date, ResourceSlot]] = []

        total_seconds = (spec.period_end - spec.period_start).total_seconds()
        if total_seconds <= 0:
            return result

        # Fast path: most specs don't cross midnight
        if spec.period_start.date() == spec.period_end.date():
            return [(spec.period_start.date(), spec.resource_usage)]

        # Slow path: split across day boundaries
        current_start = spec.period_start

        while current_start < spec.period_end:
            current_date = current_start.date()
            # Next midnight
            next_midnight = datetime.combine(
                current_date + timedelta(days=1),
                datetime.min.time(),
                tzinfo=current_start.tzinfo,
            )

            # Segment ends at next midnight or spec end, whichever is earlier
            segment_end = min(next_midnight, spec.period_end)
            segment_seconds = (segment_end - current_start).total_seconds()

            if segment_seconds > 0:
                # Proportionally allocate resource usage
                proportion = Decimal(str(segment_seconds)) / Decimal(str(total_seconds))
                segment_usage = ResourceSlot({
                    key: value * proportion for key, value in spec.resource_usage.items()
                })
                result.append((current_date, segment_usage))

            current_start = segment_end

        return result

    def _add_to_bucket_deltas(
        self,
        spec: KernelUsageRecordCreatorSpec,
        period_date: date,
        resource_usage: ResourceSlot,
        user_deltas: dict[UserUsageBucketKey, ResourceSlot],
        project_deltas: dict[ProjectUsageBucketKey, ResourceSlot],
        domain_deltas: dict[DomainUsageBucketKey, ResourceSlot],
    ) -> None:
        """Add resource usage to bucket deltas for a day.

        Args:
            spec: Original spec (for entity identifiers)
            period_date: Date of the bucket
            resource_usage: Resource usage for this day segment
            user_deltas: User deltas to update (mutated)
            project_deltas: Project deltas to update (mutated)
            domain_deltas: Domain deltas to update (mutated)
        """
        # User bucket key
        user_key = UserUsageBucketKey(
            user_uuid=spec.user_uuid,
            project_id=spec.project_id,
            domain_name=spec.domain_name,
            resource_group=spec.resource_group,
            period_date=period_date,
        )
        user_deltas[user_key] = user_deltas[user_key] + resource_usage

        # Project bucket key
        project_key = ProjectUsageBucketKey(
            project_id=spec.project_id,
            domain_name=spec.domain_name,
            resource_group=spec.resource_group,
            period_date=period_date,
        )
        project_deltas[project_key] = project_deltas[project_key] + resource_usage

        # Domain bucket key
        domain_key = DomainUsageBucketKey(
            domain_name=spec.domain_name,
            resource_group=spec.resource_group,
            period_date=period_date,
        )
        domain_deltas[domain_key] = domain_deltas[domain_key] + resource_usage

    def _prepare_kernel_usage_specs(
        self,
        kernel: KernelInfo,
        scaling_group: str,
        now: datetime,
    ) -> tuple[list[KernelUsageRecordCreatorSpec], datetime]:
        """Prepare usage record specs for a single kernel.

        Generates 5-minute slices aligned to clock boundaries.
        Partial slices are only allowed at:
        - Start: When starts_at is not on a boundary (first observation only)
        - End: When the kernel terminates at a non-boundary time

        For RUNNING kernels, only complete slices up to the last boundary are generated.

        Args:
            kernel: Kernel to process
            scaling_group: The scaling group
            now: Current time from DB

        Returns:
            Tuple of (list of specs, last_observed_at to save)
        """
        is_first_observation = kernel.lifecycle.last_observed_at is None
        is_terminated = kernel.lifecycle.terminated_at is not None

        # Determine start time
        start_time: datetime
        if is_first_observation:
            if kernel.lifecycle.starts_at is None:
                raise ValueError(f"Kernel {kernel.id} has no starts_at for first observation")
            start_time = kernel.lifecycle.starts_at
        else:
            if kernel.lifecycle.last_observed_at is None:
                raise ValueError(
                    f"Kernel {kernel.id} has no last_observed_at for non-first observation"
                )
            start_time = kernel.lifecycle.last_observed_at

        # Determine end time based on kernel status
        end_time: datetime
        if is_terminated:
            # TERMINATED: use terminated_at (allows partial end slice)
            if kernel.lifecycle.terminated_at is None:
                raise ValueError(
                    f"Kernel {kernel.id} marked as terminated but has no terminated_at"
                )
            end_time = kernel.lifecycle.terminated_at
        else:
            # RUNNING: floor to last complete boundary (no partial end slice)
            end_time = self._floor_to_boundary(now)

        log.debug(
            "[Aggregator] Kernel {}: is_first={}, is_terminated={}, "
            "start_time={}, end_time={}, now={}",
            kernel.id,
            is_first_observation,
            is_terminated,
            start_time,
            end_time,
            now,
        )

        # Validate time range
        if end_time <= start_time:
            # Not enough time has passed to generate any slices
            log.debug(
                "[Aggregator] Kernel {}: skipped (end_time <= start_time)",
                kernel.id,
            )
            return [], start_time

        # Generate 5-minute slices
        specs = self._generate_slice_specs(
            kernel=kernel,
            scaling_group=scaling_group,
            start_time=start_time,
            end_time=end_time,
        )

        log.debug(
            "[Aggregator] Kernel {}: generated {} specs",
            kernel.id,
            len(specs),
        )

        # Return end_time as the new last_observed_at
        # For RUNNING: this is floored to boundary
        # For TERMINATED: this is terminated_at
        return specs, end_time

    def _generate_slice_specs(
        self,
        kernel: KernelInfo,
        scaling_group: str,
        start_time: datetime,
        end_time: datetime,
    ) -> list[KernelUsageRecordCreatorSpec]:
        """Generate 5-minute slice specs aligned to clock boundaries.

        Slices are aligned to 5-minute clock boundaries (00:00, 00:05, 00:10, etc.).
        The caller is responsible for adjusting start_time and end_time to control
        whether partial slices are allowed.

        Args:
            kernel: Kernel info
            scaling_group: The scaling group
            start_time: Start of observation period
            end_time: End of observation period

        Returns:
            List of usage record specs for each slice
        """
        specs: list[KernelUsageRecordCreatorSpec] = []

        current_start = start_time
        while current_start < end_time:
            # Calculate next 5-minute boundary after current_start
            start_epoch = int(current_start.timestamp())
            next_boundary_epoch = (
                (start_epoch // SLICE_DURATION_SECONDS) + 1
            ) * SLICE_DURATION_SECONDS
            next_boundary = datetime.fromtimestamp(next_boundary_epoch, tz=current_start.tzinfo)

            # Slice ends at the next boundary or end_time, whichever is earlier
            current_end = min(next_boundary, end_time)
            slice_seconds = (current_end - current_start).total_seconds()

            if slice_seconds <= 0:
                break

            # Calculate resource-seconds for this slice
            resource_seconds = self._calculate_resource_seconds(
                kernel.resource.occupied_slots,
                slice_seconds,
            )

            spec = KernelUsageRecordCreatorSpec(
                kernel_id=UUID(str(kernel.id)),
                session_id=UUID(kernel.session.session_id),
                user_uuid=kernel.user_permission.user_uuid,
                project_id=kernel.user_permission.group_id,
                domain_name=kernel.user_permission.domain_name,
                resource_group=scaling_group,
                period_start=current_start,
                period_end=current_end,
                resource_usage=resource_seconds,
            )
            specs.append(spec)

            current_start = current_end

        return specs

    def _floor_to_boundary(self, dt: datetime) -> datetime:
        """Floor datetime to the previous 5-minute boundary.

        Examples:
            07:47:30 -> 07:45:00
            07:50:00 -> 07:50:00 (already on boundary)
            07:52:15 -> 07:50:00

        Args:
            dt: Datetime to floor

        Returns:
            Datetime floored to 5-minute boundary
        """
        epoch = int(dt.timestamp())
        floored_epoch = (epoch // SLICE_DURATION_SECONDS) * SLICE_DURATION_SECONDS
        return datetime.fromtimestamp(floored_epoch, tz=dt.tzinfo)

    def _calculate_resource_seconds(
        self,
        slots: ResourceSlot,
        seconds: float,
    ) -> ResourceSlot:
        """Convert resource slots to resource-seconds.

        Multiplies each resource value by the number of seconds to get
        the total resource-seconds consumed during the period.

        Args:
            slots: Resource slots (e.g., {"cpu": 2, "mem": 4096})
            seconds: Duration in seconds

        Returns:
            ResourceSlot with values in resource-seconds
        """
        return ResourceSlot({key: value * Decimal(str(seconds)) for key, value in slots.items()})



"""empty message

Revision ID: c1409ad0e8da
Revises: 22e52d03fc61, 9a91532c8534
Create Date: 2019-07-29 20:18:52.291350

"""

# revision identifiers, used by Alembic.
revision = "c1409ad0e8da"
down_revision = ("22e52d03fc61", "9a91532c8534")
branch_labels = None
depends_on = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



"""Session preparer that applies multiple preparation rules."""

import logging
import uuid
from collections.abc import Iterable
from datetime import datetime
from typing import Any

from ai.backend.common.types import (
    KernelEnqueueingConfig,
    KernelId,
    ResourceSlot,
    SessionId,
    VFolderMount,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.models.network import NetworkType
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    AllowedScalingGroup,
    KernelEnqueueData,
    SessionCreationContext,
    SessionCreationSpec,
    SessionEnqueueData,
)
from ai.backend.manager.sokovan.scheduling_controller.types import CalculatedResources

from .base import SessionPreparerRule

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class SessionPreparer:
    """
    Preparer that applies multiple preparation rules for session creation.
    """

    _rules: Iterable[SessionPreparerRule]

    # Default values for missing data
    DEFAULT_IMAGE_NAME = "ai.backend.undefined"
    DEFAULT_ARCHITECTURE = "x86_64"
    DEFAULT_REGISTRY = "localhost"

    def __init__(self, rules: Iterable[SessionPreparerRule]) -> None:
        self._rules = rules

    async def prepare(
        self,
        spec: SessionCreationSpec,
        validated_scaling_group: AllowedScalingGroup,
        context: SessionCreationContext,
        calculated_resources: CalculatedResources,
        enqueue_time: datetime,
    ) -> SessionEnqueueData:
        """
        Convert creation spec to enqueue data.

        Args:
            spec: Session creation specification
            validated_scaling_group: Validated scaling group
            context: Pre-fetched context with all required data
            calculated_resources: Pre-calculated resource information

        Returns:
            Complete SessionEnqueueData with kernels and dependencies
        """
        # Initialize preparation data
        preparation_data: dict[str, Any] = {
            "dotfile_data": context.dotfile_data,
            "vfolder_mounts": context.vfolder_mounts,
        }

        # Apply all preparation rules
        for rule in self._rules:
            log.debug(f"Applying preparation rule: {rule.name()}")
            rule.prepare(spec, context, preparation_data)

        # Generate session ID if not provided
        session_id = SessionId(uuid.uuid4())

        # Use network info from context
        scaling_group_network = context.scaling_group_network

        # Prepare environment variables
        environ = dict(spec.creation_spec.get("environ") or {})

        # Add SUDO_SESSION_ENABLED to environ if sudo session is enabled
        if spec.sudo_session_enabled:
            environ["SUDO_SESSION_ENABLED"] = "1"

        # Determine network configuration
        network_type, network_id = self._determine_network_config(spec, context)

        # Extract bootstrap script and timeout
        bootstrap_script = spec.creation_spec.get("bootstrap_script")
        timeout = spec.creation_spec.get("session_timeout")

        # Build kernel data using kernel_configs from preparation_data
        kernel_configs = preparation_data.get("kernel_configs", spec.kernel_specs)
        internal_data = preparation_data.get("internal_data", {})

        kernel_data_list = await self._prepare_kernels(
            spec,
            session_id,
            validated_scaling_group,
            kernel_configs,
            internal_data,
            context.vfolder_mounts,
            context,
            calculated_resources,
            enqueue_time,
        )

        # Collect images from kernels
        session_images = self._collect_session_images(kernel_data_list)

        # Use pre-calculated session total
        total_requested = calculated_resources.session_requested_slots

        # Build complete session data with all information
        return SessionEnqueueData(
            id=session_id,
            creation_id=spec.session_creation_id,
            name=spec.session_name,
            access_key=spec.access_key,
            user_uuid=spec.user_scope.user_uuid,
            group_id=spec.user_scope.group_id,
            domain_name=spec.user_scope.domain_name,
            scaling_group_name=validated_scaling_group.name,
            session_type=spec.session_type,
            cluster_mode=spec.cluster_mode,
            cluster_size=spec.cluster_size,
            priority=spec.priority,
            status=SessionStatus.PENDING.name,
            status_history={
                SessionStatus.PENDING.name: enqueue_time.isoformat(),
            },
            requested_slots=total_requested,
            occupying_slots=ResourceSlot(),
            vfolder_mounts=context.vfolder_mounts,
            environ=environ,
            tag=spec.session_tag,
            starts_at=spec.starts_at,
            batch_timeout=(int(spec.batch_timeout.total_seconds()) if spec.batch_timeout else None),
            callback_url=spec.callback_url,
            images=session_images,
            network_type=network_type,
            network_id=network_id,
            bootstrap_script=bootstrap_script,
            designated_agent_list=self._get_designated_agent_ids(
                spec=spec, kernel_config=kernel_configs
            ),
            use_host_network=scaling_group_network.use_host_network,
            timeout=timeout,
            kernels=kernel_data_list,
            dependencies=spec.dependency_sessions or [],
            startup_command=spec.startup_command,
        )

    def _determine_network_config(
        self,
        spec: SessionCreationSpec,
        context: SessionCreationContext,
    ) -> tuple[NetworkType | None, str | None]:
        """Determine network type and ID from spec."""
        if spec.network:
            return NetworkType.PERSISTENT, str(spec.network.id)
        if context.scaling_group_network.use_host_network:
            return NetworkType.HOST, None
        # Default to VOLATILE for multi-container or single-container sessions
        return NetworkType.VOLATILE, None

    async def _prepare_kernels(
        self,
        spec: SessionCreationSpec,
        session_id: SessionId,
        validated_scaling_group: AllowedScalingGroup,
        kernel_configs: list[KernelEnqueueingConfig],
        internal_data: dict[str, Any],
        vfolder_mounts: list[VFolderMount],
        context: SessionCreationContext,
        calculated_resources: CalculatedResources,
        enqueue_time: datetime,
    ) -> list[KernelEnqueueData]:
        """Prepare kernel enqueue data."""
        kernel_data_list = []

        for idx, kernel_config in enumerate(kernel_configs):
            kernel_id = KernelId(uuid.uuid4())

            # Get image info
            image_ref = kernel_config.get("image_ref")
            if image_ref and hasattr(image_ref, "canonical"):
                image_info = context.image_infos.get(image_ref.canonical)
            else:
                # Fallback for string image references
                image_info = None

            # Get preopen_ports from kernel config or fall back to creation_spec
            # This follows the same pattern as in the original registry.py
            preopen_ports = kernel_config.get("creation_config", {}).get("preopen_ports")
            if preopen_ports is None:
                # Fall back to creation_spec preopen_ports (applies to all kernels)
                preopen_ports = spec.creation_spec.get("preopen_ports")
            if not preopen_ports:
                preopen_ports = []
            if not isinstance(preopen_ports, list):
                preopen_ports = []

            # Use pre-calculated resources for this kernel
            if idx < len(calculated_resources.kernel_resources):
                kernel_resource = calculated_resources.kernel_resources[idx]
                requested_slots = kernel_resource.requested_slots
                resource_opts = kernel_resource.resource_opts
                log.debug(
                    "Kernel {} (idx={}, role={}) using calculated resources: {}",
                    kernel_id,
                    idx,
                    kernel_config.get("cluster_role", "unknown"),
                    requested_slots,
                )
            else:
                # Fallback if no calculated resources
                log.warning(
                    "Kernel {} (idx={}, role={}) has no calculated resources! "
                    "calculated_resources has {} entries but need idx={}",
                    kernel_id,
                    idx,
                    kernel_config.get("cluster_role", "unknown"),
                    len(calculated_resources.kernel_resources),
                    idx,
                )
                requested_slots = ResourceSlot()
                resource_opts = {}

            # Build kernel data
            kernel_data = KernelEnqueueData(
                id=kernel_id,
                session_id=session_id,
                session_creation_id=spec.session_creation_id,
                session_name=spec.session_name,
                session_type=spec.session_type,
                cluster_mode=spec.cluster_mode,
                cluster_size=spec.cluster_size,
                cluster_role=kernel_config.get("cluster_role", DEFAULT_ROLE),
                cluster_idx=kernel_config.get("cluster_idx", idx),
                local_rank=kernel_config.get("local_rank", idx),
                cluster_hostname=kernel_config.get("cluster_hostname")
                or f"{kernel_config.get('cluster_role', DEFAULT_ROLE)}{kernel_config.get('cluster_idx', idx + 1)}",
                scaling_group=validated_scaling_group.name,
                domain_name=spec.user_scope.domain_name,
                group_id=spec.user_scope.group_id,
                user_uuid=spec.user_scope.user_uuid,
                access_key=spec.access_key,
                image=image_info.canonical if image_info else self.DEFAULT_IMAGE_NAME,
                architecture=image_info.architecture if image_info else self.DEFAULT_ARCHITECTURE,
                registry=image_info.registry if image_info else self.DEFAULT_REGISTRY,
                tag=spec.session_tag,
                starts_at=spec.starts_at,
                status=KernelStatus.PENDING.name,
                status_history={
                    KernelStatus.PENDING.name: enqueue_time.isoformat(),
                },
                occupied_slots=ResourceSlot(),
                requested_slots=requested_slots,
                occupied_shares={},
                resource_opts=resource_opts,
                environ=[f"{k}={v}" for k, v in (kernel_config.get("environ", {}) or {}).items()]  # type: ignore[attr-defined]
                if kernel_config.get("environ")
                else [],
                bootstrap_script=kernel_config.get("bootstrap_script"),
                startup_command=kernel_config.get("startup_command"),
                internal_data=internal_data,
                callback_url=spec.callback_url,
                mounts=[mount.name for mount in vfolder_mounts],  # Legacy field
                vfolder_mounts=vfolder_mounts,
                preopen_ports=preopen_ports if isinstance(preopen_ports, list) else [],
                use_host_network=context.scaling_group_network.use_host_network,
                uid=kernel_config.get("uid") or context.container_user_info.uid,
                main_gid=kernel_config.get("main_gid") or context.container_user_info.main_gid,
                gids=kernel_config.get("supplementary_gids")
                or context.container_user_info.supplementary_gids,
            )

            kernel_data_list.append(kernel_data)

        return kernel_data_list

    def _get_designated_agent_ids(
        self,
        spec: SessionCreationSpec,
        kernel_config: list[KernelEnqueueingConfig],
    ) -> list[str] | None:
        """Get pre-assigned agent for a kernel."""
        # Check if agent is specified in kernel config
        designated_agents = set()
        for kernel_cfg in kernel_config:
            if agent_id := kernel_cfg.get("agent"):
                designated_agents.add(str(agent_id))
        if designated_agents:
            return list(designated_agents)
        return spec.designated_agent_list

    def _collect_session_images(
        self,
        kernel_data_list: list[KernelEnqueueData],
    ) -> list[str]:
        """Collect unique images from kernels, with main kernel first."""
        session_images: list[str] = []

        for kernel in kernel_data_list:
            if kernel.image not in session_images:
                if kernel.cluster_role == DEFAULT_ROLE:
                    # Main kernel image goes first
                    session_images.insert(0, kernel.image)
                else:
                    session_images.append(kernel.image)

        return session_images
